# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========

import asyncio
import os
from typing import TYPE_CHECKING, Dict, List, Optional, Union

if TYPE_CHECKING:
    from aci.types.app_configurations import AppConfiguration
    from aci.types.apps import AppBasic, AppDetails
    from aci.types.linked_accounts import LinkedAccount

from camel.logger import get_logger
from camel.toolkits import FunctionTool
from camel.toolkits.base import BaseToolkit
from camel.utils import api_keys_required, dependencies_required

logger = get_logger(__name__)


@api_keys_required(
    [
        (None, 'ACI_API_KEY'),
    ]
)
class ACIToolkit(BaseToolkit):
    r"""A toolkit for interacting with the ACI API."""

    @dependencies_required('aci')
    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        linked_account_owner_id: Optional[str] = None,
        timeout: Optional[float] = None,
    ) -> None:
        r"""Initialize the ACI toolkit.

        Args:
            api_key (Optional[str]): The API key for authentication.
                (default: :obj:`None`)
            base_url (Optional[str]): The base URL for the ACI API.
                (default: :obj:`None`)
            linked_account_owner_id (Optional[str]): ID of the owner of the
                linked account, e.g., "johndoe"
                (default: :obj:`None`)
            timeout (Optional[float]): Request timeout.
                (default: :obj:`None`)
        """
        from aci import ACI

        super().__init__(timeout)

        self._api_key = api_key or os.getenv("ACI_API_KEY")
        self._base_url = base_url or os.getenv("ACI_BASE_URL")
        self.client = ACI(api_key=self._api_key, base_url=self._base_url)
        self.linked_account_owner_id = linked_account_owner_id

    def search_tool(
        self,
        intent: Optional[str] = None,
        allowed_app_only: bool = True,
        include_functions: bool = False,
        categories: Optional[List[str]] = None,
        limit: Optional[int] = 10,
        offset: Optional[int] = 0,
    ) -> Union[List["AppBasic"], str]:
        r"""Search for apps based on intent.

        Args:
            intent (Optional[str]): Search results will be sorted by relevance
                to this intent.
                (default: :obj:`None`)
            allowed_app_only (bool): If true, only return apps that
                are allowed by the agent/accessor, identified by the api key.
                (default: :obj:`True`)
            include_functions (bool): If true, include functions
                (name and description) in the search results.
                (default: :obj:`False`)
            categories (Optional[List[str]]): List of categories to filter the
                search results. Defaults to an empty list.
                (default: :obj:`None`)
            limit (Optional[int]): Maximum number of results to return.
                (default: :obj:`10`)
            offset (Optional[int]): Offset for pagination.
                (default: :obj:`0`)

        Returns:
            Optional[List[AppBasic]]: List of matching apps if successful,
                error message otherwise.
        """
        try:
            apps = self.client.apps.search(
                intent=intent,
                allowed_apps_only=allowed_app_only,
                include_functions=include_functions,
                categories=categories,
                limit=limit,
                offset=offset,
            )
            return apps
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def list_configured_apps(
        self,
        app_names: Optional[List[str]] = None,
        limit: Optional[int] = 10,
        offset: Optional[int] = 0,
    ) -> Union[List["AppConfiguration"], str]:
        r"""List all configured apps.

        Args:
            app_names (Optional[List[str]]): List of app names to filter the
                results. (default: :obj:`None`)
            limit (Optional[int]): Maximum number of results to return.
                (default: :obj:`10`)
            offset (Optional[int]): Offset for pagination. (default: :obj:`0`)

        Returns:
            Union[List[AppConfiguration], str]: List of configured apps if
                successful, error message otherwise.
        """
        try:
            apps = self.client.app_configurations.list(
                app_names=app_names, limit=limit, offset=offset
            )
            return apps
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def configure_app(self, app_name: str) -> Union[Dict, str]:
        r"""Configure an app with specified authentication type.

        Args:
            app_name (str): Name of the app to configure.

        Returns:
            Union[Dict, str]: Configuration result or error message.
        """
        from aci.types.enums import SecurityScheme

        try:
            app_details = self.get_app_details(app_name)
            if app_details and app_details.security_schemes[0] == "api_key":
                security_scheme = SecurityScheme.API_KEY
            elif app_details and app_details.security_schemes[0] == "oauth2":
                security_scheme = SecurityScheme.OAUTH2
            else:
                security_scheme = SecurityScheme.NO_AUTH
            configuration = self.client.app_configurations.create(
                app_name=app_name, security_scheme=security_scheme
            )
            return configuration
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def get_app_configuration(
        self, app_name: str
    ) -> Union["AppConfiguration", str]:
        r"""Get app configuration by app name.

        Args:
            app_name (str): Name of the app to get configuration for.

        Returns:
            Union[AppConfiguration, str]: App configuration if successful,
                error message otherwise.
        """
        try:
            app = self.client.app_configurations.get(app_name=app_name)
            return app
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def delete_app(self, app_name: str) -> Optional[str]:
        r"""Delete an app configuration.

        Args:
            app_name (str): Name of the app to delete.

        Returns:
            Optional[str]: None if successful, error message otherwise.
        """
        try:
            self.client.app_configurations.delete(app_name=app_name)
            return None
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def link_account(
        self,
        app_name: str,
    ) -> Union["LinkedAccount", str]:
        r"""Link an account to a configured app.

        Args:
            app_name (str): Name of the app to link the account to.

        Returns:
            Union[LinkedAccount, str]: LinkedAccount object if successful,
                error message otherwise.
        """
        from aci.types.enums import SecurityScheme

        try:
            security_scheme = self.client.app_configurations.get(
                app_name=app_name
            ).security_scheme

            if security_scheme == SecurityScheme.API_KEY:
                return self.client.linked_accounts.link(
                    app_name=app_name,
                    linked_account_owner_id=self.linked_account_owner_id,
                    security_scheme=security_scheme,
                    api_key=self._api_key,
                )
            else:
                return self.client.linked_accounts.link(
                    app_name=app_name,
                    linked_account_owner_id=self.linked_account_owner_id,
                    security_scheme=security_scheme,
                )
        except Exception as e:
            logger.error(f"Error linking account: {e!s}")
            return str(e)

    def get_app_details(self, app_name: str) -> "AppDetails":
        r"""Get details of an app.

        Args:
            app_name (str): Name of the app to get details for.

        Returns:
            AppDetails: App details.
        """
        app = self.client.apps.get(app_name=app_name)
        return app

    def get_linked_accounts(
        self, app_name: str
    ) -> Union[List["LinkedAccount"], str]:
        r"""List all linked accounts for a specific app.

        Args:
            app_name (str): Name of the app to get linked accounts for.

        Returns:
            Union[List[LinkedAccount], str]: List of linked accounts if
                successful, error message otherwise.
        """
        try:
            accounts = self.client.linked_accounts.list(app_name=app_name)
            return accounts
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def enable_linked_account(
        self, linked_account_id: str
    ) -> Union["LinkedAccount", str]:
        r"""Enable a linked account.

        Args:
            linked_account_id (str): ID of the linked account to enable.

        Returns:
            Union[LinkedAccount, str]: Linked account if successful, error
                message otherwise.
        """
        try:
            linked_account = self.client.linked_accounts.enable(
                linked_account_id=linked_account_id
            )
            return linked_account
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def disable_linked_account(
        self, linked_account_id: str
    ) -> Union["LinkedAccount", str]:
        r"""Disable a linked account.

        Args:
            linked_account_id (str): ID of the linked account to disable.

        Returns:
            Union[LinkedAccount, str]: The updated linked account if
                successful, error message otherwise.
        """
        try:
            linked_account = self.client.linked_accounts.disable(
                linked_account_id=linked_account_id
            )
            return linked_account
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def delete_linked_account(self, linked_account_id: str) -> str:
        r"""Delete a linked account.

        Args:
            linked_account_id (str): ID of the linked account to delete.

        Returns:
            str: Success message if successful, error message otherwise.
        """
        try:
            self.client.linked_accounts.delete(
                linked_account_id=linked_account_id
            )
            return (
                f"linked_account_id: {linked_account_id} deleted successfully"
            )
        except Exception as e:
            logger.error(f"Error: {e}")
            return str(e)

    def function_definition(self, func_name: str) -> Dict:
        r"""Get the function definition for an app.

        Args:
            app_name (str): Name of the app to get function definition for

        Returns:
            Dict: Function definition dictionary.
        """
        return self.client.functions.get_definition(func_name)

    def search_function(
        self,
        app_names: Optional[List[str]] = None,
        intent: Optional[str] = None,
        allowed_apps_only: bool = True,
        limit: Optional[int] = 10,
        offset: Optional[int] = 0,
    ) -> List[Dict]:
        r"""Search for functions based on intent.

        Args:
            app_names (Optional[List[str]]): List of app names to filter the
                search results. (default: :obj:`None`)
            intent (Optional[str]): The search query/intent.
                (default: :obj:`None`)
            allowed_apps_only (bool): If true, only return
                functions from allowed apps. (default: :obj:`True`)
            limit (Optional[int]): Maximum number of results to return.
                (default: :obj:`10`)
            offset (Optional[int]): Offset for pagination.
                (default: :obj:`0`)

        Returns:
            List[Dict]: List of matching functions
        """
        return self.client.functions.search(
            app_names=app_names,
            intent=intent,
            allowed_apps_only=allowed_apps_only,
            limit=limit,
            offset=offset,
        )

    def execute_function(
        self,
        function_name: str,
        function_arguments: Dict,
        linked_account_owner_id: str,
        allowed_apps_only: bool = False,
    ) -> Dict:
        r"""Execute a function call.

        Args:
            function_name (str): Name of the function to execute.
            function_arguments (Dict): Arguments to pass to the function.
            linked_account_owner_id (str): To specify the end-user (account
                owner) on behalf of whom you want to execute functions
                You need to first link corresponding account with the same
                owner id in the ACI dashboard (https://platform.aci.dev).
            allowed_apps_only (bool): If true, only returns functions/apps
                that are allowed to be used by the agent/accessor, identified
                by the api key. (default: :obj:`False`)

        Returns:
            Dict: Result of the function execution
        """
        result = self.client.handle_function_call(
            function_name,
            function_arguments,
            linked_account_owner_id,
            allowed_apps_only,
        )
        return result

    async def aexecute_function(
        self,
        function_name: str,
        function_arguments: Dict,
        linked_account_owner_id: str,
        allowed_apps_only: bool = False,
    ) -> Dict:
        r"""Execute a function call asynchronously.

        Args:
            function_name (str): Name of the function to execute.
            function_arguments (Dict): Arguments to pass to the function.
            linked_account_owner_id (str): To specify the end-user (account
                owner) on behalf of whom you want to execute functions
                You need to first link corresponding account with the same
                owner id in the ACI dashboard (https://platform.aci.dev).
            allowed_apps_only (bool): If true, only returns functions/apps
                that are allowed to be used by the agent/accessor, identified
                by the api key. (default: :obj:`False`)

        Returns:
            Dict: Result of the function execution
        """
        result = await asyncio.to_thread(
            self.client.handle_function_call,
            function_name,
            function_arguments,
            linked_account_owner_id,
            allowed_apps_only,
        )
        return result

    def get_tools(self) -> List[FunctionTool]:
        r"""Get a list of tools (functions) available in the configured apps.

        Returns:
            List[FunctionTool]: List of FunctionTool objects representing
                available functions
        """
        _configure_app = [
            app.app_name  # type: ignore[union-attr]
            for app in self.list_configured_apps() or []
        ]
        _all_function = self.search_function(app_names=_configure_app)
        tools = [
            FunctionTool(self.search_tool),
            FunctionTool(self.list_configured_apps),
            FunctionTool(self.configure_app),
            FunctionTool(self.get_app_configuration),
            FunctionTool(self.delete_app),
            FunctionTool(self.link_account),
            FunctionTool(self.get_app_details),
            FunctionTool(self.get_linked_accounts),
            FunctionTool(self.enable_linked_account),
            FunctionTool(self.disable_linked_account),
            FunctionTool(self.delete_linked_account),
            FunctionTool(self.function_definition),
            FunctionTool(self.search_function),
            FunctionTool(self.execute_function),
            FunctionTool(self.aexecute_function),
        ]

        for function in _all_function:
            schema = self.client.functions.get_definition(
                function['function']['name']
            )

            def dummy_func(*, schema=schema, **kwargs):
                return self.execute_function(
                    function_name=schema['function']['name'],
                    function_arguments=kwargs,
                    linked_account_owner_id=self.linked_account_owner_id,
                )

            async def async_dummy_func(*, schema=schema, **kwargs):
                return await self.aexecute_function(
                    function_name=schema['function']['name'],
                    function_arguments=kwargs,
                    linked_account_owner_id=self.linked_account_owner_id,
                )

            # Add async_call method to the sync function for compatibility
            dummy_func.async_call = async_dummy_func  # type: ignore[attr-defined]

            tool = FunctionTool(
                func=dummy_func,
                openai_tool_schema=schema,
            )
            tools.append(tool)
        return tools



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
import json
import os
import tempfile
from typing import Any, List, Optional

from camel.datahubs.base import BaseDatasetManager
from camel.datahubs.models import Record
from camel.logger import get_logger
from camel.types import HuggingFaceRepoType
from camel.utils import api_keys_required, dependencies_required

logger = get_logger(__name__)


class HuggingFaceDatasetManager(BaseDatasetManager):
    r"""A dataset manager for Hugging Face datasets. This class provides
    methods to create, add, update, delete, and list records in a dataset on
    the Hugging Face Hub.

    Args:
        token (str): The Hugging Face API token. If not provided, the token
            will be read from the environment variable `HF_TOKEN`.
    """

    @api_keys_required(
        [
            ("token", "HF_TOKEN"),
        ]
    )
    @dependencies_required('huggingface_hub')
    def __init__(self, token: Optional[str] = None):
        from huggingface_hub import HfApi

        self._api_key = token or os.getenv("HF_TOKEN")
        self.api = HfApi(token=self._api_key)

    def create_dataset_card(
        self,
        dataset_name: str,
        description: str,
        license: Optional[str] = None,
        version: Optional[str] = None,
        tags: Optional[List[str]] = None,
        authors: Optional[List[str]] = None,
        size_category: Optional[List[str]] = None,
        language: Optional[List[str]] = None,
        task_categories: Optional[List[str]] = None,
        content: Optional[str] = None,
    ) -> None:
        r"""Creates and uploads a dataset card to the Hugging Face Hub in YAML
            format.

        Args:
            dataset_name (str): The name of the dataset.
            description (str): A description of the dataset.
            license (str): The license of the dataset. (default: :obj:`None`)
            version (str): The version of the dataset. (default: :obj:`None`)
            tags (list): A list of tags for the dataset.(default: :obj:`None`)
            authors (list): A list of authors of the dataset. (default:
                :obj:`None`)
            size_category (list): A size category for the dataset. (default:
                :obj:`None`)
            language (list): A list of languages the dataset is in. (default:
                :obj:`None`)
            task_categories (list): A list of task categories. (default:
                :obj:`None`)
            content (str): Custom markdown content that the user wants to add
                to the dataset card. (default: :obj:`None`)
        """
        import yaml

        metadata = {
            "license": license,
            "authors": authors,
            "task_categories": task_categories,
            "language": language,
            "tags": tags,
            "pretty_name": dataset_name,
            "size_categories": size_category,
            "version": version,
            "description": description,
        }

        # Remove keys with None values
        metadata = {k: v for k, v in metadata.items() if v}

        card_content = (
            "---\n"
            + yaml.dump(metadata, default_flow_style=False, allow_unicode=True)
            + "\n---"
        )

        if content:
            card_content += f"\n\n# Additional Information\n{content}\n"

        self._upload_file(
            file_content=card_content,
            dataset_name=dataset_name,
            filepath="README.md",
            file_type="md",
        )

    def create_dataset(
        self, name: str, private: bool = False, **kwargs: Any
    ) -> str:
        r"""Creates a new dataset on the Hugging Face Hub.

        Args:
            name (str): The name of the dataset.
            private (bool): Whether the dataset should be private. defaults to
                False.
            kwargs (Any): Additional keyword arguments.

        Returns:
            str: The URL of the created dataset.
        """
        from huggingface_hub.errors import RepositoryNotFoundError

        try:
            self.api.repo_info(
                repo_id=name,
                repo_type=HuggingFaceRepoType.DATASET.value,
                **kwargs,
            )
        except RepositoryNotFoundError:
            self.api.create_repo(
                repo_id=name,
                repo_type=HuggingFaceRepoType.DATASET.value,
                private=private,
            )

        return f"https://huggingface.co/datasets/{name}"

    def list_datasets(
        self, username: str, limit: int = 100, **kwargs: Any
    ) -> List[str]:
        r"""Lists all datasets for the current user.

        Args:
            username (str): The username of the user whose datasets to list.
            limit (int): The maximum number of datasets to list.
                (default: :obj:`100`)
            kwargs (Any): Additional keyword arguments.

        Returns:
            List[str]: A list of dataset ids.
        """
        try:
            return [
                dataset.id
                for dataset in self.api.list_datasets(
                    author=username, limit=limit, **kwargs
                )
            ]
        except Exception as e:
            logger.error(f"Error listing datasets: {e}")
            return []

    def delete_dataset(self, dataset_name: str, **kwargs: Any) -> None:
        r"""Deletes a dataset from the Hugging Face Hub.

        Args:
            dataset_name (str): The name of the dataset to delete.
            kwargs (Any): Additional keyword arguments.
        """
        try:
            self.api.delete_repo(
                repo_id=dataset_name,
                repo_type=HuggingFaceRepoType.DATASET.value,
                **kwargs,
            )
            logger.info(f"Dataset '{dataset_name}' deleted successfully.")
        except Exception as e:
            logger.error(f"Error deleting dataset '{dataset_name}': {e}")
            raise

    def add_records(
        self,
        dataset_name: str,
        records: List[Record],
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> None:
        r"""Adds records to a dataset on the Hugging Face Hub.

        Args:
            dataset_name (str): The name of the dataset.
            records (List[Record]): A list of records to add to the dataset.
            filepath (str): The path to the file containing the records.
            kwargs (Any): Additional keyword arguments.

        Raises:
            ValueError: If the dataset already has a records file.
        """
        existing_records = self._download_records(
            dataset_name=dataset_name, filepath=filepath, **kwargs
        )

        if existing_records:
            raise ValueError(
                f"Dataset '{filepath}' already exists. "
                f"Use `update_records` to modify."
            )

        self._upload_records(
            records=records,
            dataset_name=dataset_name,
            filepath=filepath,
            **kwargs,
        )

    def update_records(
        self,
        dataset_name: str,
        records: List[Record],
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> None:
        r"""Updates records in a dataset on the Hugging Face Hub.

        Args:
            dataset_name (str): The name of the dataset.
            records (List[Record]): A list of records to update in the dataset.
            filepath (str): The path to the file containing the records.
            kwargs (Any): Additional keyword arguments.

        Raises:
            ValueError: If the dataset does not have an existing file to update
                records in.
        """
        existing_records = self._download_records(
            dataset_name=dataset_name, filepath=filepath, **kwargs
        )

        if not existing_records:
            logger.warning(
                f"Dataset '{dataset_name}' does not have existing "
                "records. Adding new records."
            )
            self._upload_records(
                records=records,
                dataset_name=dataset_name,
                filepath=filepath,
                **kwargs,
            )
            return

        old_dict = {record.id: record for record in existing_records}
        new_dict = {record.id: record for record in records}
        merged_dict = old_dict.copy()
        merged_dict.update(new_dict)

        self._upload_records(
            records=list(merged_dict.values()),
            dataset_name=dataset_name,
            filepath=filepath,
            **kwargs,
        )

    def delete_record(
        self,
        dataset_name: str,
        record_id: str,
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> None:
        r"""Deletes a record from the dataset.

        Args:
            dataset_name (str): The name of the dataset.
            record_id (str): The ID of the record to delete.
            filepath (str): The path to the file containing the records.
            kwargs (Any): Additional keyword arguments.

        Raises:
            ValueError: If the dataset does not have an existing file to delete
                records from.
        """
        existing_records = self._download_records(
            dataset_name=dataset_name, filepath=filepath, **kwargs
        )

        if not existing_records:
            raise ValueError(
                f"Dataset '{dataset_name}' does not have an existing file to "
                f"delete records from."
            )

        filtered_records = [
            record for record in existing_records if record.id != record_id
        ]

        self._upload_records(
            records=filtered_records,
            dataset_name=dataset_name,
            filepath=filepath,
            **kwargs,
        )

    def list_records(
        self,
        dataset_name: str,
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> List[Record]:
        r"""Lists all records in a dataset.

        Args:
            dataset_name (str): The name of the dataset.
            filepath (str): The path to the file containing the records.
            kwargs (Any): Additional keyword arguments.

        Returns:
            List[Record]: A list of records in the dataset.
        """
        return self._download_records(
            dataset_name=dataset_name, filepath=filepath, **kwargs
        )

    def _download_records(
        self, dataset_name: str, filepath: str, **kwargs: Any
    ) -> List[Record]:
        from huggingface_hub import hf_hub_download
        from huggingface_hub.errors import EntryNotFoundError

        try:
            downloaded_file_path = hf_hub_download(
                repo_id=dataset_name,
                filename=filepath,
                repo_type=HuggingFaceRepoType.DATASET.value,
                token=self._api_key,
                **kwargs,
            )

            with open(downloaded_file_path, "r") as f:
                records_data = json.load(f)

            return [Record(**record) for record in records_data]
        except EntryNotFoundError:
            logger.info(f"No records found for dataset '{dataset_name}'.")
            return []
        except Exception as e:
            logger.error(f"Error downloading or processing records: {e}")
            raise e

    def _upload_records(
        self,
        records: List[Record],
        dataset_name: str,
        filepath: str,
        **kwargs: Any,
    ):
        with tempfile.NamedTemporaryFile(
            delete=False, mode="w", newline="", encoding="utf-8"
        ) as f:
            json.dump(
                [
                    record.model_dump(exclude_defaults=True)
                    for record in records
                ],
                f,
                ensure_ascii=False,
            )
            temp_file_path = f.name

        try:
            self.api.upload_file(
                path_or_fileobj=temp_file_path,
                path_in_repo=filepath,
                repo_id=dataset_name,
                repo_type=HuggingFaceRepoType.DATASET.value,
                **kwargs,
            )
        except Exception as e:
            logger.error(f"Error uploading records file: {e}")
            raise
        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def _upload_file(
        self,
        file_content: str,
        dataset_name: str,
        filepath: str,
        file_type: str = "json",
        **kwargs: Any,
    ):
        with tempfile.NamedTemporaryFile(
            mode="w", delete=False, suffix=f".{file_type}"
        ) as f:
            if file_type == "json":
                if isinstance(file_content, str):
                    try:
                        json_content = json.loads(file_content)
                    except json.JSONDecodeError:
                        raise ValueError(
                            "Invalid JSON string provided for file_content."
                        )
                else:
                    try:
                        json.dumps(file_content, ensure_ascii=False)
                        json_content = file_content
                    except (TypeError, ValueError):
                        raise ValueError(
                            "file_content is not JSON serializable."
                        )

                json.dump(json_content, f, ensure_ascii=False)
            elif file_type == "md" or file_type == "txt":
                f.write(file_content)
            else:
                raise ValueError(f"Unsupported file type: {file_type}")

            temp_file_path = f.name

        try:
            self.api.upload_file(
                path_or_fileobj=temp_file_path,
                path_in_repo=filepath,
                repo_id=dataset_name,
                repo_type=HuggingFaceRepoType.DATASET.value,
                **kwargs,
            )
            logger.info(f"File uploaded successfully: {filepath}")
        except Exception as e:
            logger.error(f"Error uploading file: {e}")
            raise

        if os.path.exists(temp_file_path):
            os.remove(temp_file_path)



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========

from .agent_memories import (
    ChatHistoryMemory,
    LongtermAgentMemory,
    VectorDBMemory,
)
from .base import AgentMemory, BaseContextCreator, MemoryBlock
from .blocks.chat_history_block import ChatHistoryBlock
from .blocks.vectordb_block import VectorDBBlock
from .context_creators.score_based import ScoreBasedContextCreator
from .records import ContextRecord, MemoryRecord

__all__ = [
    'MemoryRecord',
    'ContextRecord',
    'MemoryBlock',
    "AgentMemory",
    'BaseContextCreator',
    'ScoreBasedContextCreator',
    'ChatHistoryMemory',
    'VectorDBMemory',
    'ChatHistoryBlock',
    'VectorDBBlock',
    'LongtermAgentMemory',
]



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
from __future__ import annotations

from typing import List, Optional

from camel.configs.base_config import BaseConfig


class FunctionGemmaConfig(BaseConfig):
    r"""Defines the parameters for generating completions using FunctionGemma
    via Ollama's native API.

    FunctionGemma uses a custom chat template format for function calling
    that differs from OpenAI's format. This config is used with Ollama's
    /api/generate endpoint.

    Reference: https://github.com/ollama/ollama/blob/main/docs/api.md

    Args:
        temperature (float, optional): Sampling temperature to use, between
            :obj:`0` and :obj:`2`. Higher values make the output more random,
            while lower values make it more focused and deterministic.
            (default: :obj:`None`)
        top_p (float, optional): An alternative to sampling with temperature,
            called nucleus sampling, where the model considers the results of
            the tokens with top_p probability mass. (default: :obj:`0.95`)
        top_k (int, optional): Limits the next token selection to the K most
            probable tokens. (default: :obj:`64`)
        num_predict (int, optional): Maximum number of tokens to generate.
            (default: :obj:`None`)
        stop (list, optional): Sequences where the model will stop generating
            further tokens. (default: :obj:`None`)
        seed (int, optional): Random seed for reproducibility.
            (default: :obj:`None`)
    """

    temperature: Optional[float] = None
    top_p: Optional[float] = 0.95
    top_k: Optional[int] = 64
    num_predict: Optional[int] = None
    stop: Optional[List[str]] = None
    seed: Optional[int] = None


FUNCTION_GEMMA_API_PARAMS = {
    param for param in FunctionGemmaConfig.model_fields.keys()
}



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========

from typing import List, Optional

from .mcp_toolkit import MCPToolkit


class PlaywrightMCPToolkit(MCPToolkit):
    r"""PlaywrightMCPToolkit provides an interface for interacting with web
    browsers using the Playwright automation library through the Model Context
    Protocol (MCP).

    Attributes:
        timeout (Optional[float]): Connection timeout in seconds.
            (default: :obj:`None`)
        additional_args (Optional[List[str]]): Additional command-line
            arguments to pass to the Playwright MCP server. For example,
            `["--cdp-endpoint=http://localhost:9222"]`.
            (default: :obj:`None`)

    Note:
        Currently only supports asynchronous operation mode.
    """

    def __init__(
        self,
        timeout: Optional[float] = None,
        additional_args: Optional[List[str]] = None,
    ) -> None:
        r"""Initializes the PlaywrightMCPToolkit.

        Args:
            timeout (Optional[float]): Connection timeout in seconds.
                (default: :obj:`None`)
            additional_args (Optional[List[str]]): Additional command-line
                arguments to pass to the Playwright MCP server. For example,
                `["--cdp-endpoint=http://localhost:9222"]`.
                (default: :obj:`None`)
        """
        # Create config for Playwright MCP server
        config_dict = {
            "mcpServers": {
                "playwright": {
                    "command": "npx",
                    "args": ["@playwright/mcp@latest"]
                    + (additional_args or []),
                }
            }
        }

        # Initialize parent MCPToolkit with Playwright configuration
        super().__init__(config_dict=config_dict, timeout=timeout)



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
from abc import ABC, abstractmethod
from typing import Any, Callable

DEFAULT_TOP_K_RESULTS = 1


def _query_unimplemented(self, *input: Any) -> None:
    r"""Defines the query behavior performed at every call.

    Query the results. Subclasses should implement this
        method according to their specific needs.

    It should be overridden by all subclasses.

    .. note::
        Although the recipe for forward pass needs to be defined within
        this function, one should call the :class:`BaseRetriever` instance
        afterwards instead of this since the former takes care of running the
        registered hooks while the latter silently ignores them.
    """
    raise NotImplementedError(
        f"Retriever [{type(self).__name__}] is missing the required"
        " \"query\" function"
    )


def _process_unimplemented(self, *input: Any) -> None:
    r"""Defines the process behavior performed at every call.

    Processes content from a file or URL, divides it into chunks by
        using `Unstructured IO`,then stored internally. This method must be
        called before executing queries with the retriever.

    Should be overridden by all subclasses.

    .. note::
        Although the recipe for forward pass needs to be defined within
        this function, one should call the :class:`BaseRetriever` instance
        afterwards instead of this since the former takes care of running the
        registered hooks while the latter silently ignores them.
    """
    raise NotImplementedError(
        f"Retriever [{type(self).__name__}] is missing the required "
        "\"process\" function"
    )


class BaseRetriever(ABC):
    r"""Abstract base class for implementing various types of information
    retrievers.
    """

    @abstractmethod
    def __init__(self) -> None:
        pass

    process: Callable[..., Any] = _process_unimplemented
    query: Callable[..., Any] = _query_unimplemented



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
import json
import os
import time
import uuid
from typing import Any, Dict, List, Optional, Type, Union

import httpx
from openai import AsyncOpenAI, AsyncStream, OpenAI, Stream
from pydantic import BaseModel

from camel.configs import (
    SambaCloudAPIConfig,
)
from camel.messages import OpenAIMessage
from camel.models import BaseModelBackend
from camel.types import (
    ChatCompletion,
    ChatCompletionChunk,
    CompletionUsage,
    ModelType,
)
from camel.utils import (
    BaseTokenCounter,
    OpenAITokenCounter,
    api_keys_required,
    update_current_observation,
)

try:
    if os.getenv("AGENTOPS_API_KEY") is not None:
        from agentops import LLMEvent, record
    else:
        raise ImportError
except (ImportError, AttributeError):
    LLMEvent = None

if os.environ.get("LANGFUSE_ENABLED", "False").lower() == "true":
    try:
        from langfuse.decorators import observe
    except ImportError:
        from camel.utils import observe
else:
    from camel.utils import observe


class SambaModel(BaseModelBackend):
    r"""SambaNova service interface.

    Args:
        model_type (Union[ModelType, str]): Model for which a SambaNova backend
            is created. Supported models via SambaNova Cloud:
            `https://community.sambanova.ai/t/supported-models/193`.
            Supported models via SambaVerse API is listed in
            `https://sambaverse.sambanova.ai/models`.
        model_config_dict (Optional[Dict[str, Any]], optional): A dictionary
            that will be fed into:obj:`openai.ChatCompletion.create()`. If
            :obj:`None`, :obj:`SambaCloudAPIConfig().as_dict()` will be used.
            (default: :obj:`None`)
        api_key (Optional[str], optional): The API key for authenticating
            with the SambaNova service. (default: :obj:`None`)
        url (Optional[str], optional): The url to the SambaNova service.
            Current support SambaVerse API:
            :obj:`"https://sambaverse.sambanova.ai/api/predict"` and
            SambaNova Cloud:
            :obj:`"https://api.sambanova.ai/v1"` (default: :obj:`https://api.
            sambanova.ai/v1`)
        token_counter (Optional[BaseTokenCounter], optional): Token counter to
            use for the model. If not provided, :obj:`OpenAITokenCounter(
            ModelType.GPT_4O_MINI)` will be used.
        timeout (Optional[float], optional): The timeout value in seconds for
            API calls. If not provided, will fall back to the MODEL_TIMEOUT
            environment variable or default to 180 seconds.
            (default: :obj:`None`)
        max_retries (int, optional): Maximum number of retries for API calls.
            (default: :obj:`3`)
        client (Optional[Any], optional): A custom synchronous
            OpenAI-compatible client instance. If provided, this client will
            be used instead of creating a new one. Only applicable when using
            SambaNova Cloud API. (default: :obj:`None`)
        async_client (Optional[Any], optional): A custom asynchronous
            OpenAI-compatible client instance. If provided, this client will
            be used instead of creating a new one. Only applicable when using
            SambaNova Cloud API. (default: :obj:`None`)
        **kwargs (Any): Additional arguments to pass to the client
            initialization. Ignored if custom clients are provided.
    """

    @api_keys_required(
        [
            ("api_key", 'SAMBA_API_KEY'),
        ]
    )
    def __init__(
        self,
        model_type: Union[ModelType, str],
        model_config_dict: Optional[Dict[str, Any]] = None,
        api_key: Optional[str] = None,
        url: Optional[str] = None,
        token_counter: Optional[BaseTokenCounter] = None,
        timeout: Optional[float] = None,
        max_retries: int = 3,
        client: Optional[Any] = None,
        async_client: Optional[Any] = None,
        **kwargs: Any,
    ) -> None:
        if model_config_dict is None:
            model_config_dict = SambaCloudAPIConfig().as_dict()
        api_key = api_key or os.environ.get("SAMBA_API_KEY")
        url = url or os.environ.get(
            "SAMBA_API_BASE_URL",
            "https://api.sambanova.ai/v1",
        )
        timeout = timeout or float(os.environ.get("MODEL_TIMEOUT", 180))
        super().__init__(
            model_type,
            model_config_dict,
            api_key,
            url,
            token_counter,
            timeout,
            max_retries,
        )

        # Only create clients for Cloud API mode
        if self._url == "https://api.sambanova.ai/v1":
            # Use custom clients if provided, otherwise create new ones
            if client is not None:
                self._client = client
            else:
                self._client = OpenAI(
                    timeout=self._timeout,
                    max_retries=self._max_retries,
                    base_url=self._url,
                    api_key=self._api_key,
                    **kwargs,
                )

            if async_client is not None:
                self._async_client = async_client
            else:
                self._async_client = AsyncOpenAI(
                    timeout=self._timeout,
                    max_retries=self._max_retries,
                    base_url=self._url,
                    api_key=self._api_key,
                    **kwargs,
                )

    @property
    def token_counter(self) -> BaseTokenCounter:
        r"""Initialize the token counter for the model backend.

        Returns:
            BaseTokenCounter: The token counter following the model's
                tokenization style.
        """
        if not self._token_counter:
            self._token_counter = OpenAITokenCounter(ModelType.GPT_4O_MINI)
        return self._token_counter

    @observe(as_type="generation")
    async def _arun(  # type: ignore[misc]
        self,
        messages: List[OpenAIMessage],
        response_format: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        r"""Runs SambaNova's service.

        Args:
            messages (List[OpenAIMessage]): Message list with the chat history
                in OpenAI API format.

        Returns:
            Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
                `ChatCompletion` in the non-stream mode, or
                `AsyncStream[ChatCompletionChunk]` in the stream mode.
        """

        update_current_observation(
            input={
                "messages": messages,
                "tools": tools,
            },
            model=str(self.model_type),
            model_parameters=self.model_config_dict,
        )

        self._log_and_trace()

        if "tools" in self.model_config_dict:
            del self.model_config_dict["tools"]
        if self.model_config_dict.get("stream") is True:
            return await self._arun_streaming(messages)
        else:
            response = await self._arun_non_streaming(messages)
            update_current_observation(
                usage=response.usage,
            )
            return response

    @observe(as_type="generation")
    def _run(  # type: ignore[misc]
        self,
        messages: List[OpenAIMessage],
        response_format: Optional[Type[BaseModel]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
    ) -> Union[ChatCompletion, Stream[ChatCompletionChunk]]:
        r"""Runs SambaNova's service.

        Args:
            messages (List[OpenAIMessage]): Message list with the chat history
                in OpenAI API format.

        Returns:
            Union[ChatCompletion, Stream[ChatCompletionChunk]]:
                `ChatCompletion` in the non-stream mode, or
                `Stream[ChatCompletionChunk]` in the stream mode.
        """
        update_current_observation(
            input={
                "messages": messages,
                "tools": tools,
            },
            model=str(self.model_type),
            model_parameters=self.model_config_dict,
        )
        self._log_and_trace()

        if "tools" in self.model_config_dict:
            del self.model_config_dict["tools"]
        if self.model_config_dict.get("stream") is True:
            return self._run_streaming(messages)
        else:
            response = self._run_non_streaming(messages)
            update_current_observation(
                usage=response.usage,
            )
            return response

    def _run_streaming(
        self, messages: List[OpenAIMessage]
    ) -> Stream[ChatCompletionChunk]:
        r"""Handles streaming inference with SambaNova's API.

        Args:
            messages (List[OpenAIMessage]): A list of messages representing the
                chat history in OpenAI API format.

        Returns:
            Stream[ChatCompletionChunk]: A generator yielding
                `ChatCompletionChunk` objects as they are received from the
                API.

        Raises:
            RuntimeError: If the HTTP request fails.
            ValueError: If the API doesn't support stream mode.
        """
        # Handle SambaNova's Cloud API
        if self._url == "https://api.sambanova.ai/v1":
            response = self._client.chat.completions.create(
                messages=messages,
                model=self.model_type,
                **self.model_config_dict,
            )

            # Add AgentOps LLM Event tracking
            if LLMEvent:
                llm_event = LLMEvent(
                    thread_id=response.id,
                    prompt=" ".join(
                        [message.get("content") for message in messages]  # type: ignore[misc]
                    ),
                    prompt_tokens=response.usage.prompt_tokens,  # type: ignore[union-attr]
                    completion=response.choices[0].message.content,
                    completion_tokens=response.usage.completion_tokens,  # type: ignore[union-attr]
                    model=self.model_type,
                )
                record(llm_event)

            return response

        elif self._url == "https://sambaverse.sambanova.ai/api/predict":
            raise ValueError(
                "https://sambaverse.sambanova.ai/api/predict doesn't support"
                " stream mode"
            )
        raise RuntimeError(f"Unknown URL: {self._url}")

    def _run_non_streaming(
        self, messages: List[OpenAIMessage]
    ) -> ChatCompletion:
        r"""Handles non-streaming inference with SambaNova's API.

        Args:
            messages (List[OpenAIMessage]): A list of messages representing the
                message in OpenAI API format.

        Returns:
            ChatCompletion: A `ChatCompletion` object containing the complete
                response from the API.

        Raises:
            RuntimeError: If the HTTP request fails.
            ValueError: If the JSON response cannot be decoded or is missing
                expected data.
        """
        # Handle SambaNova's Cloud API
        if self._url == "https://api.sambanova.ai/v1":
            response = self._client.chat.completions.create(
                messages=messages,
                model=self.model_type,
                **self.model_config_dict,
            )

            # Add AgentOps LLM Event tracking
            if LLMEvent:
                llm_event = LLMEvent(
                    thread_id=response.id,
                    prompt=" ".join(
                        [message.get("content") for message in messages]  # type: ignore[misc]
                    ),
                    prompt_tokens=response.usage.prompt_tokens,  # type: ignore[union-attr]
                    completion=response.choices[0].message.content,
                    completion_tokens=response.usage.completion_tokens,  # type: ignore[union-attr]
                    model=self.model_type,
                )
                record(llm_event)

            return response

        # Handle SambaNova's Sambaverse API
        else:
            headers = {
                "Content-Type": "application/json",
                "key": str(self._api_key),
                "modelName": self.model_type,
            }

            data = {
                "instance": json.dumps(
                    {
                        "conversation_id": str(uuid.uuid4()),
                        "messages": messages,
                    },
                    ensure_ascii=False,
                ),
                "params": {
                    "do_sample": {"type": "bool", "value": "true"},
                    "max_tokens_to_generate": {
                        "type": "int",
                        "value": str(self.model_config_dict.get("max_tokens")),
                    },
                    "process_prompt": {"type": "bool", "value": "true"},
                    "repetition_penalty": {
                        "type": "float",
                        "value": str(
                            self.model_config_dict.get("repetition_penalty")
                        ),
                    },
                    "return_token_count_only": {
                        "type": "bool",
                        "value": "false",
                    },
                    "select_expert": {
                        "type": "str",
                        "value": self.model_type.split('/')[1],
                    },
                    "stop_sequences": {
                        "type": "str",
                        "value": self.model_config_dict.get("stop_sequences"),
                    },
                    "temperature": {
                        "type": "float",
                        "value": str(
                            self.model_config_dict.get("temperature")
                        ),
                    },
                    "top_k": {
                        "type": "int",
                        "value": str(self.model_config_dict.get("top_k")),
                    },
                    "top_p": {
                        "type": "float",
                        "value": str(self.model_config_dict.get("top_p")),
                    },
                },
            }

            try:
                # Send the request and handle the response
                with httpx.Client() as client:
                    response = client.post(
                        self._url,  # type: ignore[arg-type]
                        headers=headers,
                        json=data,
                    )

                raw_text = response.text
                # Split the string into two dictionaries
                dicts = raw_text.split('}\n{')

                # Keep only the last dictionary
                last_dict = '{' + dicts[-1]

                # Parse the dictionary
                last_dict = json.loads(last_dict)
                return self._sambaverse_to_openai_response(last_dict)  # type: ignore[arg-type]

            except httpx.HTTPStatusError:
                raise RuntimeError(f"HTTP request failed: {raw_text}")

    def _sambaverse_to_openai_response(
        self, samba_response: Dict[str, Any]
    ) -> ChatCompletion:
        r"""Converts SambaVerse API response into an OpenAI-compatible
        response.

        Args:
            samba_response (Dict[str, Any]): A dictionary representing
                responses from the SambaVerse API.

        Returns:
            ChatCompletion: A `ChatCompletion` object constructed from the
                aggregated response data.
        """
        choices = [
            dict(
                index=0,
                message={
                    "role": 'assistant',
                    "content": samba_response['result']['responses'][0][
                        'completion'
                    ],
                },
                finish_reason=samba_response['result']['responses'][0][
                    'stop_reason'
                ],
            )
        ]

        obj = ChatCompletion.construct(
            id=None,
            choices=choices,
            created=int(time.time()),
            model=self.model_type,
            object="chat.completion",
            # SambaVerse API only provide `total_tokens`
            usage=CompletionUsage(
                completion_tokens=0,
                prompt_tokens=0,
                total_tokens=int(
                    samba_response['result']['responses'][0][
                        'total_tokens_count'
                    ]
                ),
            ),
        )

        return obj

    @property
    def stream(self) -> bool:
        r"""Returns whether the model is in stream mode, which sends partial
        results each time.

        Returns:
            bool: Whether the model is in stream mode.
        """
        return self.model_config_dict.get('stream', False)

    async def _arun_streaming(
        self, messages: List[OpenAIMessage]
    ) -> AsyncStream[ChatCompletionChunk]:
        r"""Handles streaming inference with SambaNova's API.

        Args:
            messages (List[OpenAIMessage]): A list of messages representing the
                chat history in OpenAI API format.

        Returns:
            AsyncStream[ChatCompletionChunk]: A generator yielding
                `ChatCompletionChunk` objects as they are received from the
                API.

        Raises:
            RuntimeError: If the HTTP request fails.
            ValueError: If the API doesn't support stream mode.
        """
        # Handle SambaNova's Cloud API
        if self._url == "https://api.sambanova.ai/v1":
            response = await self._async_client.chat.completions.create(
                messages=messages,
                model=self.model_type,
                **self.model_config_dict,
            )

            # Add AgentOps LLM Event tracking
            if LLMEvent:
                llm_event = LLMEvent(
                    thread_id=response.id,
                    prompt=" ".join(
                        [message.get("content") for message in messages]  # type: ignore[misc]
                    ),
                    prompt_tokens=response.usage.prompt_tokens,  # type: ignore[union-attr]
                    completion=response.choices[0].message.content,
                    completion_tokens=response.usage.completion_tokens,  # type: ignore[union-attr]
                    model=self.model_type,
                )
                record(llm_event)

            return response

        elif self._url == "https://sambaverse.sambanova.ai/api/predict":
            raise ValueError(
                "https://sambaverse.sambanova.ai/api/predict doesn't support"
                " stream mode"
            )
        raise RuntimeError(f"Unknown URL: {self._url}")

    async def _arun_non_streaming(
        self, messages: List[OpenAIMessage]
    ) -> ChatCompletion:
        r"""Handles non-streaming inference with SambaNova's API.

        Args:
            messages (List[OpenAIMessage]): A list of messages representing the
                message in OpenAI API format.

        Returns:
            ChatCompletion: A `ChatCompletion` object containing the complete
                response from the API.

        Raises:
            RuntimeError: If the HTTP request fails.
            ValueError: If the JSON response cannot be decoded or is missing
                expected data.
        """
        # Handle SambaNova's Cloud API
        if self._url == "https://api.sambanova.ai/v1":
            response = await self._async_client.chat.completions.create(
                messages=messages,
                model=self.model_type,
                **self.model_config_dict,
            )

            # Add AgentOps LLM Event tracking
            if LLMEvent:
                llm_event = LLMEvent(
                    thread_id=response.id,
                    prompt=" ".join(
                        [message.get("content") for message in messages]  # type: ignore[misc]
                    ),
                    prompt_tokens=response.usage.prompt_tokens,  # type: ignore[union-attr]
                    completion=response.choices[0].message.content,
                    completion_tokens=response.usage.completion_tokens,  # type: ignore[union-attr]
                    model=self.model_type,
                )
                record(llm_event)

            return response

        # Handle SambaNova's Sambaverse API
        else:
            headers = {
                "Content-Type": "application/json",
                "key": str(self._api_key),
                "modelName": self.model_type,
            }

            data = {
                "instance": json.dumps(
                    {
                        "conversation_id": str(uuid.uuid4()),
                        "messages": messages,
                    },
                    ensure_ascii=False,
                ),
                "params": {
                    "do_sample": {"type": "bool", "value": "true"},
                    "max_tokens_to_generate": {
                        "type": "int",
                        "value": str(self.model_config_dict.get("max_tokens")),
                    },
                    "process_prompt": {"type": "bool", "value": "true"},
                    "repetition_penalty": {
                        "type": "float",
                        "value": str(
                            self.model_config_dict.get("repetition_penalty")
                        ),
                    },
                    "return_token_count_only": {
                        "type": "bool",
                        "value": "false",
                    },
                    "select_expert": {
                        "type": "str",
                        "value": self.model_type.split("/")[1],
                    },
                    "stop_sequences": {
                        "type": "str",
                        "value": self.model_config_dict.get("stop_sequences"),
                    },
                    "temperature": {
                        "type": "float",
                        "value": str(
                            self.model_config_dict.get("temperature")
                        ),
                    },
                    "top_k": {
                        "type": "int",
                        "value": str(self.model_config_dict.get("top_k")),
                    },
                    "top_p": {
                        "type": "float",
                        "value": str(self.model_config_dict.get("top_p")),
                    },
                },
            }

            try:
                # Send the request and handle the response
                with httpx.Client() as client:
                    response = client.post(
                        self._url,  # type: ignore[arg-type]
                        headers=headers,
                        json=data,
                    )

                raw_text = response.text
                # Split the string into two dictionaries
                dicts = raw_text.split("}\n{")

                # Keep only the last dictionary
                last_dict = "{" + dicts[-1]

                # Parse the dictionary
                last_dict = json.loads(last_dict)
                return self._sambaverse_to_openai_response(last_dict)  # type: ignore[arg-type]

            except httpx.HTTPStatusError:
                raise RuntimeError(f"HTTP request failed: {raw_text}")



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
from abc import ABC, abstractmethod
from typing import Any, List

from camel.datahubs.models import Record


class BaseDatasetManager(ABC):
    r"""Abstract base class for dataset managers."""

    @abstractmethod
    def create_dataset(self, name: str, **kwargs: Any) -> str:
        r"""Creates a new dataset.

        Args:
            name (str): The name of the dataset.
            kwargs (Any): Additional keyword arguments.

        Returns:
            str: The URL of the created dataset.
        """
        pass

    @abstractmethod
    def list_datasets(
        self, username: str, limit: int = 100, **kwargs: Any
    ) -> List[str]:
        r"""Lists all datasets for the current user.

        Args:
            username (str): The username of the user whose datasets to list.
            limit (int): The maximum number of datasets to list.
                (default::obj:`100`)
            kwargs (Any): Additional keyword arguments.

        Returns:
            List[str]: A list of dataset ids.
        """
        pass

    @abstractmethod
    def delete_dataset(self, dataset_name: str, **kwargs: Any) -> None:
        r"""Deletes a dataset.

        Args:
            dataset_name (str): The name of the dataset to delete.
            kwargs (Any): Additional keyword arguments.
        """
        pass

    @abstractmethod
    def add_records(
        self,
        dataset_name: str,
        records: List[Record],
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> None:
        r"""Adds records to a dataset.

        Args:
            dataset_name (str): The name of the dataset.
            records (List[Record]): A list of records to add to the dataset.
            filepath (str): The path to the file containing the records.
                (default::obj:`"records/records.json"`)
            kwargs (Any): Additional keyword arguments.
        """
        pass

    @abstractmethod
    def update_records(
        self,
        dataset_name: str,
        records: List[Record],
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> None:
        r"""Updates records in a dataset.

        Args:
            dataset_name (str): The name of the dataset.
            records (List[Record]): A list of records to update in the dataset.
            filepath (str): The path to the file containing the records.
                (default::obj:`"records/records.json"`)
            kwargs (Any): Additional keyword arguments.
        """
        pass

    @abstractmethod
    def list_records(
        self,
        dataset_name: str,
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> List[Record]:
        r"""Lists records in a dataset.

        Args:
            dataset_name (str): The name of the dataset.
            filepath (str): The path to the file containing the records.
                (default::obj:`"records/records.json"`)
            kwargs (Any): Additional keyword arguments.
        """
        pass

    # New method for record deletion
    @abstractmethod
    def delete_record(
        self,
        dataset_name: str,
        record_id: str,
        filepath: str = "records/records.json",
        **kwargs: Any,
    ) -> None:
        r"""Deletes a record from the dataset.

        Args:
            dataset_name (str): The name of the dataset.
            record_id (str): The ID of the record to delete.
            filepath (str): The path to the file containing the records.
                (default::obj:`"records/records.json"`)
            kwargs (Any): Additional keyword arguments.
        """
        pass



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========

import asyncio
import contextlib
import datetime
import json
import os
import subprocess
import time
import uuid
from contextvars import ContextVar
from functools import wraps
from typing import TYPE_CHECKING, Any, Dict, List, Optional

if TYPE_CHECKING:
    import websockets
else:
    try:
        import websockets
    except ImportError:
        websockets = None

from camel.logger import get_logger
from camel.utils.tool_result import ToolResult

from .installer import check_and_install_dependencies

logger = get_logger(__name__)

# Context variable to track if we're inside a high-level action
_in_high_level_action: ContextVar[bool] = ContextVar(
    '_in_high_level_action', default=False
)


def _create_memory_aware_error(base_msg: str) -> str:
    import psutil

    mem = psutil.virtual_memory()
    if mem.available < 1024**3:
        return (
            f"{base_msg} "
            f"(likely due to insufficient memory). "
            f"Available memory: {mem.available / 1024**3:.2f}GB "
            f"({mem.percent}% used)"
        )
    return base_msg


async def _cleanup_process_and_tasks(process, log_reader_task, ts_log_file):
    if process:
        with contextlib.suppress(ProcessLookupError, Exception):
            process.kill()
        with contextlib.suppress(Exception):
            process.wait(timeout=2)

    if log_reader_task and not log_reader_task.done():
        log_reader_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await log_reader_task

    if ts_log_file:
        with contextlib.suppress(Exception):
            ts_log_file.close()


def action_logger(func):
    """Decorator to add logging to action methods.

    Skips logging if already inside a high-level action to avoid
    logging internal calls.
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        # Skip logging if we're already inside a high-level action
        if _in_high_level_action.get():
            return await func(self, *args, **kwargs)

        action_name = func.__name__
        start_time = time.time()

        inputs = {
            "args": args,
            "kwargs": kwargs,
        }

        try:
            result = await func(self, *args, **kwargs)
            execution_time = time.time() - start_time

            page_load_time = None
            if isinstance(result, dict) and 'page_load_time_ms' in result:
                page_load_time = result['page_load_time_ms'] / 1000.0

            await self._log_action(
                action_name=action_name,
                inputs=inputs,
                outputs=result,
                execution_time=execution_time,
                page_load_time=page_load_time,
            )

            return result

        except Exception as e:
            execution_time = time.time() - start_time
            error_msg = f"{type(e).__name__}: {e!s}"

            await self._log_action(
                action_name=action_name,
                inputs=inputs,
                outputs=None,
                execution_time=execution_time,
                error=error_msg,
            )

            raise

    return wrapper


def high_level_action(func):
    """Decorator for high-level actions that should suppress low-level logging.

    When a function is decorated with this, all low-level action_logger
    decorated functions called within it will skip logging. This decorator
    itself will log the high-level action.
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        action_name = func.__name__
        start_time = time.time()

        inputs = {
            "args": args,
            "kwargs": kwargs,
        }

        # Set the context variable to indicate we're in a high-level action
        token = _in_high_level_action.set(True)
        try:
            result = await func(self, *args, **kwargs)
            execution_time = time.time() - start_time

            # Log the high-level action
            if hasattr(self, '_get_ws_wrapper'):
                # This is a HybridBrowserToolkit instance
                ws_wrapper = await self._get_ws_wrapper()
                await ws_wrapper._log_action(
                    action_name=action_name,
                    inputs=inputs,
                    outputs=result,
                    execution_time=execution_time,
                    page_load_time=None,
                )

            return result

        except Exception as e:
            execution_time = time.time() - start_time
            error_msg = f"{type(e).__name__}: {e!s}"

            # Log the error
            if hasattr(self, '_get_ws_wrapper'):
                ws_wrapper = await self._get_ws_wrapper()
                await ws_wrapper._log_action(
                    action_name=action_name,
                    inputs=inputs,
                    outputs=None,
                    execution_time=execution_time,
                    error=error_msg,
                )

            raise
        finally:
            # Reset the context variable
            _in_high_level_action.reset(token)

    return wrapper


class WebSocketBrowserWrapper:
    """Python wrapper for the TypeScript hybrid browser
    toolkit implementation using WebSocket."""

    def __init__(self, config: Optional[Dict[str, Any]] = None):
        """Initialize the wrapper.

        Args:
            config: Configuration dictionary for the browser toolkit
        """
        if websockets is None:
            raise ImportError(
                "websockets package is required for WebSocket communication. "
                "Install with: pip install websockets"
            )

        self.config = config or {}
        self.ts_dir = os.path.join(os.path.dirname(__file__), 'ts')
        self.process: Optional[subprocess.Popen] = None
        self.websocket = None
        self.server_port = None
        self._send_lock = asyncio.Lock()
        self._request_timeout = self.config.get(
            'requestTimeout', 60
        )  # seconds
        self._receive_task = None
        self._pending_responses: Dict[str, asyncio.Future[Dict[str, Any]]] = {}
        self._browser_opened = False
        self._server_ready_future = None

        self.browser_log_to_file = (config or {}).get(
            'browser_log_to_file', False
        )
        self.log_dir = (config or {}).get('log_dir', 'browser_log')
        self.session_id = (config or {}).get('session_id', 'default')
        self.log_file_path: Optional[str] = None
        self.log_buffer: List[Dict[str, Any]] = []
        self.ts_log_file_path: Optional[str] = None
        self.ts_log_file = None
        self._log_reader_task = None

        if self.browser_log_to_file:
            log_dir = self.log_dir if self.log_dir else "browser_log"
            os.makedirs(log_dir, exist_ok=True)
            timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            self.log_file_path = os.path.join(
                log_dir,
                f"hybrid_browser_toolkit_ws_{timestamp}_{self.session_id}.log",
            )
            self.ts_log_file_path = os.path.join(
                log_dir,
                f"typescript_console_{timestamp}_{self.session_id}.log",
            )

    async def __aenter__(self):
        """Async context manager entry."""
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.stop()

    async def _cleanup_existing_processes(self):
        """Clean up any existing Node.js WebSocket server processes."""
        import psutil

        cleaned_count = 0
        for proc in psutil.process_iter(['pid', 'name', 'cmdline']):
            try:
                if (
                    proc.info['name']
                    and 'node' in proc.info['name'].lower()
                    and proc.info['cmdline']
                    and any(
                        'websocket-server.js' in arg
                        for arg in proc.info['cmdline']
                    )
                ):
                    if any(self.ts_dir in arg for arg in proc.info['cmdline']):
                        logger.warning(
                            f"Found existing WebSocket server process "
                            f"(PID: {proc.info['pid']}). "
                            f"Terminating it to prevent conflicts."
                        )
                        proc.terminate()
                        try:
                            proc.wait(timeout=3)
                        except psutil.TimeoutExpired:
                            proc.kill()
                        cleaned_count += 1
            except (
                psutil.NoSuchProcess,
                psutil.AccessDenied,
                psutil.ZombieProcess,
            ):
                pass

        if cleaned_count > 0:
            logger.warning(
                f"Cleaned up {cleaned_count} existing WebSocket server "
                f"process(es). This may have been caused by improper "
                f"shutdown in previous sessions."
            )
            await asyncio.sleep(0.5)

    async def start(self):
        """Start the WebSocket server and connect to it."""
        await self._cleanup_existing_processes()

        npm_cmd, node_cmd = await check_and_install_dependencies(self.ts_dir)

        import platform

        use_shell = platform.system() == 'Windows'

        self.process = subprocess.Popen(
            [node_cmd, 'websocket-server.js'],
            cwd=self.ts_dir,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            text=True,
            encoding='utf-8',
            bufsize=1,
            shell=use_shell,
        )

        self._server_ready_future = asyncio.get_running_loop().create_future()

        self._log_reader_task = asyncio.create_task(
            self._read_and_log_output()
        )

        if self.browser_log_to_file and self.ts_log_file_path:
            logger.info(
                f"TypeScript console logs will be written to: "
                f"{self.ts_log_file_path}"
            )

        server_ready = False
        timeout = 10

        try:
            await asyncio.wait_for(self._server_ready_future, timeout=timeout)
            server_ready = True
        except asyncio.TimeoutError:
            server_ready = False

        if not server_ready:
            await _cleanup_process_and_tasks(
                self.process,
                self._log_reader_task,
                getattr(self, 'ts_log_file', None),
            )
            self.ts_log_file = None
            self.process = None

            error_msg = _create_memory_aware_error(
                "WebSocket server failed to start within timeout"
            )
            raise RuntimeError(error_msg)

        max_retries = 3
        retry_delays = [1, 2, 4]

        for attempt in range(max_retries):
            try:
                connect_timeout = 10.0 + (attempt * 5.0)

                logger.info(
                    f"Attempting to connect to WebSocket server "
                    f"(attempt {attempt + 1}/{max_retries}, "
                    f"timeout: {connect_timeout}s)"
                )

                self.websocket = await asyncio.wait_for(
                    websockets.connect(
                        f"ws://localhost:{self.server_port}",
                        ping_interval=30,
                        ping_timeout=10,
                        max_size=50 * 1024 * 1024,
                    ),
                    timeout=connect_timeout,
                )
                logger.info("Connected to WebSocket server")
                break

            except asyncio.TimeoutError:
                if attempt < max_retries - 1:
                    delay = retry_delays[attempt]
                    logger.warning(
                        f"WebSocket handshake timeout "
                        f"(attempt {attempt + 1}/{max_retries}). "
                        f"Retrying in {delay} seconds..."
                    )
                    await asyncio.sleep(delay)
                else:
                    raise RuntimeError(
                        f"Failed to connect to WebSocket server after "
                        f"{max_retries} attempts: Handshake timeout"
                    )

            except Exception as e:
                if attempt < max_retries - 1 and "timed out" in str(e).lower():
                    delay = retry_delays[attempt]
                    logger.warning(
                        f"WebSocket connection failed "
                        f"(attempt {attempt + 1}/{max_retries}): {e}. "
                        f"Retrying in {delay} seconds..."
                    )
                    await asyncio.sleep(delay)
                else:
                    break

        if not self.websocket:
            await _cleanup_process_and_tasks(
                self.process,
                self._log_reader_task,
                getattr(self, 'ts_log_file', None),
            )
            self.ts_log_file = None
            self.process = None

            error_msg = _create_memory_aware_error(
                "Failed to connect to WebSocket server after multiple attempts"
            )
            raise RuntimeError(error_msg)

        self._receive_task = asyncio.create_task(self._receive_loop())

        await self._send_command('init', self.config)

        if self.config.get('cdpUrl'):
            self._browser_opened = True

    async def stop(self):
        """Stop the WebSocket connection and server."""
        if self.websocket:
            with contextlib.suppress(asyncio.TimeoutError, Exception):
                await asyncio.wait_for(
                    self._send_command('shutdown', {}),
                    timeout=2.0,
                )

            with contextlib.suppress(Exception):
                await self.websocket.close()
            self.websocket = None

        self._browser_opened = False

        # Gracefully stop the Node process before cancelling the log reader
        if self.process:
            try:
                # give the process a short grace period to exit after shutdown
                self.process.wait(timeout=2)
            except subprocess.TimeoutExpired:
                try:
                    self.process.terminate()
                    self.process.wait(timeout=3)
                except subprocess.TimeoutExpired:
                    with contextlib.suppress(ProcessLookupError, Exception):
                        self.process.kill()
                        self.process.wait()
                except Exception as e:
                    logger.warning(f"Error terminating process: {e}")
            except Exception as e:
                logger.warning(f"Error waiting for process: {e}")

        # Now cancel background tasks (reader won't block on readline)
        tasks_to_cancel = [
            ('_receive_task', self._receive_task),
            ('_log_reader_task', self._log_reader_task),
        ]
        for _, task in tasks_to_cancel:
            if task and not task.done():
                task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await task

        # Close TS log file if open
        if getattr(self, 'ts_log_file', None):
            with contextlib.suppress(Exception):
                self.ts_log_file.close()
            self.ts_log_file = None

        # Ensure process handle cleared
        self.process = None

    async def disconnect_only(self):
        """Disconnect WebSocket and stop server without closing the browser.

        This is useful for CDP mode where the browser should remain open.
        """
        if self.websocket:
            with contextlib.suppress(Exception):
                await self.websocket.close()
            self.websocket = None

        self._browser_opened = False

        # Stop the Node process
        if self.process:
            try:
                # Send SIGTERM to gracefully shutdown
                self.process.terminate()
                self.process.wait(timeout=3)
            except subprocess.TimeoutExpired:
                # Force kill if needed
                with contextlib.suppress(ProcessLookupError, Exception):
                    self.process.kill()
                    self.process.wait()
            except Exception as e:
                logger.warning(f"Error terminating process: {e}")

        # Cancel background tasks
        tasks_to_cancel = [
            ('_receive_task', self._receive_task),
            ('_log_reader_task', self._log_reader_task),
        ]
        for _, task in tasks_to_cancel:
            if task and not task.done():
                task.cancel()
                with contextlib.suppress(asyncio.CancelledError):
                    await task

        # Close TS log file if open
        if getattr(self, 'ts_log_file', None):
            with contextlib.suppress(Exception):
                self.ts_log_file.close()
            self.ts_log_file = None

        # Ensure process handle cleared
        self.process = None

        logger.info("WebSocket disconnected without closing browser")

    async def _log_action(
        self,
        action_name: str,
        inputs: Dict[str, Any],
        outputs: Any,
        execution_time: float,
        page_load_time: Optional[float] = None,
        error: Optional[str] = None,
    ) -> None:
        """Log action details with comprehensive
        information including detailed timing breakdown."""
        if not self.browser_log_to_file or not self.log_file_path:
            return

        # Create log entry
        log_entry = {
            "timestamp": datetime.datetime.now().isoformat(),
            "session_id": self.session_id,
            "action": action_name,
            "execution_time_ms": round(execution_time * 1000, 2),
            "inputs": inputs,
        }

        if error:
            log_entry["error"] = error
        else:
            # Handle ToolResult objects for JSON serialization
            if hasattr(outputs, 'text') and hasattr(outputs, 'images'):
                # This is a ToolResult object
                log_entry["outputs"] = {
                    "text": outputs.text,
                    "images_count": len(outputs.images)
                    if outputs.images
                    else 0,
                }
            else:
                log_entry["outputs"] = outputs

        if page_load_time is not None:
            log_entry["page_load_time_ms"] = round(page_load_time * 1000, 2)

        # Write to log file
        try:
            with open(self.log_file_path, 'a', encoding='utf-8') as f:
                f.write(
                    json.dumps(log_entry, ensure_ascii=False, indent=2) + '\n'
                )
        except Exception as e:
            logger.error(f"Failed to write to log file: {e}")

    async def _receive_loop(self):
        r"""Background task to receive messages from WebSocket."""
        try:
            while self.websocket:
                try:
                    response_data = await self.websocket.recv()
                    response = json.loads(response_data)

                    message_id = response.get('id')
                    if message_id and message_id in self._pending_responses:
                        # Set the result for the waiting coroutine
                        future = self._pending_responses.pop(message_id)
                        if not future.done():
                            future.set_result(response)
                    else:
                        # Log unexpected messages
                        logger.warning(
                            f"Received unexpected message: {response}"
                        )

                except asyncio.CancelledError:
                    break
                except Exception as e:
                    # Check if it's a normal WebSocket close
                    if isinstance(e, websockets.exceptions.ConnectionClosed):
                        if e.code == 1000:  # Normal closure
                            logger.debug(f"WebSocket closed normally: {e}")
                        else:
                            logger.warning(
                                f"WebSocket closed with code {e.code}: {e}"
                            )
                    else:
                        logger.error(f"Error in receive loop: {e}")
                    # Notify all pending futures of the error
                    for future in self._pending_responses.values():
                        if not future.done():
                            future.set_exception(e)
                    self._pending_responses.clear()
                    break
        finally:
            logger.debug("Receive loop terminated")

    async def _ensure_connection(self) -> None:
        """Ensure WebSocket connection is alive."""
        if not self.websocket:
            error_msg = _create_memory_aware_error("WebSocket not connected")
            raise RuntimeError(error_msg)

        # Check if connection is still alive
        try:
            # Send a ping and wait for the corresponding pong (bounded wait)
            pong_waiter = await self.websocket.ping()
            await asyncio.wait_for(pong_waiter, timeout=5.0)
        except Exception as e:
            logger.warning(f"WebSocket ping failed: {e}")
            self.websocket = None

            error_msg = _create_memory_aware_error("WebSocket connection lost")
            raise RuntimeError(error_msg)

    async def _send_command(
        self, command: str, params: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Send a command to the WebSocket server and get response."""
        await self._ensure_connection()

        # Process params to ensure refs have 'e' prefix
        params = self._process_refs_in_params(params)

        message_id = str(uuid.uuid4())
        message = {'id': message_id, 'command': command, 'params': params}

        # Create a future for this message
        loop = asyncio.get_running_loop()
        future: asyncio.Future[Dict[str, Any]] = loop.create_future()
        self._pending_responses[message_id] = future

        try:
            # Use lock only for sending to prevent interleaved messages
            async with self._send_lock:
                if self.websocket is None:
                    raise RuntimeError("WebSocket connection not established")
                await self.websocket.send(json.dumps(message))

            # Wait for response (no lock needed, handled by background
            # receiver)
            try:
                response = await asyncio.wait_for(
                    future, timeout=self._request_timeout
                )

                if not response.get('success'):
                    raise RuntimeError(
                        f"Command failed: {response.get('error')}"
                    )
                return response['result']

            except asyncio.TimeoutError:
                # Remove from pending if timeout
                self._pending_responses.pop(message_id, None)
                # Special handling for shutdown command
                if command == 'shutdown':
                    logger.debug(
                        "Shutdown command timeout is expected - "
                        "server may have closed before responding"
                    )
                    # Return a success response for shutdown
                    return {
                        'message': 'Browser shutdown (no response received)'
                    }
                raise RuntimeError(
                    f"Timeout waiting for response to command: {command}"
                )

        except Exception as e:
            # Clean up the pending response
            self._pending_responses.pop(message_id, None)

            # Check if it's a connection closed error
            if (
                "close frame" in str(e)
                or "connection closed" in str(e).lower()
            ):
                # Special handling for shutdown command
                if command == 'shutdown':
                    logger.debug(
                        f"Connection closed during shutdown (expected): {e}"
                    )
                    return {'message': 'Browser shutdown (connection closed)'}
                logger.error(f"WebSocket connection closed unexpectedly: {e}")
                # Mark connection as closed
                self.websocket = None
                raise RuntimeError(
                    f"WebSocket connection lost "
                    f"during {command} operation: {e}"
                )
            else:
                logger.error(f"WebSocket communication error: {e}")
                raise

    # Browser action methods
    @action_logger
    async def open_browser(
        self, start_url: Optional[str] = None
    ) -> Dict[str, Any]:
        """Open browser."""
        response = await self._send_command(
            'open_browser', {'startUrl': start_url}
        )
        self._browser_opened = True
        return response

    @action_logger
    async def close_browser(self) -> str:
        """Close browser."""
        response = await self._send_command('close_browser', {})
        self._browser_opened = False
        return response['message']

    @action_logger
    async def visit_page(self, url: str) -> Dict[str, Any]:
        """Visit a page.

        In non-CDP mode, automatically opens browser if not already open.
        """
        if not self._browser_opened:
            is_cdp_mode = bool(self.config.get('cdpUrl'))

            if not is_cdp_mode:
                logger.info(
                    "Browser not open, automatically opening browser..."
                )
                await self.open_browser()

        response = await self._send_command('visit_page', {'url': url})
        return response

    @action_logger
    async def get_page_snapshot(self, viewport_limit: bool = False) -> str:
        """Get page snapshot."""
        response = await self._send_command(
            'get_page_snapshot', {'viewport_limit': viewport_limit}
        )
        # The backend returns the snapshot string directly,
        # not wrapped in an object
        if isinstance(response, str):
            return response
        # Fallback if wrapped in an object
        return response.get('snapshot', '')

    @action_logger
    async def get_snapshot_for_ai(self) -> Dict[str, Any]:
        """Get snapshot for AI with element details."""
        response = await self._send_command('get_snapshot_for_ai', {})
        return response

    @action_logger
    async def get_som_screenshot(self) -> ToolResult:
        """Get screenshot."""
        logger.info("Requesting screenshot via WebSocket...")
        start_time = time.time()

        response = await self._send_command('get_som_screenshot', {})

        end_time = time.time()
        logger.info(f"Screenshot completed in {end_time - start_time:.2f}s")

        return ToolResult(text=response['text'], images=response['images'])

    def _ensure_ref_prefix(self, ref: str) -> str:
        """Ensure ref has proper prefix"""
        if not ref:
            return ref

        # If ref is purely numeric, add 'e' prefix for main frame
        if ref.isdigit():
            return f'e{ref}'

        return ref

    def _process_refs_in_params(
        self, params: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Process parameters to ensure all refs have 'e' prefix."""
        if not params:
            return params

        # Create a copy to avoid modifying the original
        processed = params.copy()

        # Handle direct ref parameters
        if 'ref' in processed:
            processed['ref'] = self._ensure_ref_prefix(processed['ref'])

        # Handle from_ref and to_ref for drag operations
        if 'from_ref' in processed:
            processed['from_ref'] = self._ensure_ref_prefix(
                processed['from_ref']
            )
        if 'to_ref' in processed:
            processed['to_ref'] = self._ensure_ref_prefix(processed['to_ref'])

        # Handle inputs array for type_multiple
        if 'inputs' in processed and isinstance(processed['inputs'], list):
            processed_inputs = []
            for input_item in processed['inputs']:
                if isinstance(input_item, dict) and 'ref' in input_item:
                    processed_input = input_item.copy()
                    processed_input['ref'] = self._ensure_ref_prefix(
                        input_item['ref']
                    )
                    processed_inputs.append(processed_input)
                else:
                    processed_inputs.append(input_item)
            processed['inputs'] = processed_inputs

        return processed

    @action_logger
    async def click(self, ref: str) -> Dict[str, Any]:
        """Click an element."""
        response = await self._send_command('click', {'ref': ref})
        return response

    @action_logger
    async def type(self, ref: str, text: str) -> Dict[str, Any]:
        """Type text into an element."""
        response = await self._send_command('type', {'ref': ref, 'text': text})
        # Log the response for debugging
        logger.debug(f"Type response for ref {ref}: {response}")
        return response

    @action_logger
    async def type_multiple(
        self, inputs: List[Dict[str, str]]
    ) -> Dict[str, Any]:
        """Type text into multiple elements."""
        response = await self._send_command('type', {'inputs': inputs})
        return response

    @action_logger
    async def select(self, ref: str, value: str) -> Dict[str, Any]:
        """Select an option."""
        response = await self._send_command(
            'select', {'ref': ref, 'value': value}
        )
        return response

    @action_logger
    async def scroll(self, direction: str, amount: int) -> Dict[str, Any]:
        """Scroll the page."""
        response = await self._send_command(
            'scroll', {'direction': direction, 'amount': amount}
        )
        return response

    @action_logger
    async def enter(self) -> Dict[str, Any]:
        """Press enter."""
        response = await self._send_command('enter', {})
        return response

    @action_logger
    async def mouse_control(
        self, control: str, x: float, y: float
    ) -> Dict[str, Any]:
        """Control the mouse to interact with browser with x, y coordinates."""
        response = await self._send_command(
            'mouse_control', {'control': control, 'x': x, 'y': y}
        )
        return response

    @action_logger
    async def mouse_drag(self, from_ref: str, to_ref: str) -> Dict[str, Any]:
        """Control the mouse to drag and drop in the browser using ref IDs."""
        response = await self._send_command(
            'mouse_drag',
            {'from_ref': from_ref, 'to_ref': to_ref},
        )
        return response

    @action_logger
    async def press_key(self, keys: List[str]) -> Dict[str, Any]:
        """Press key and key combinations."""
        response = await self._send_command('press_key', {'keys': keys})
        return response

    @action_logger
    async def back(self) -> Dict[str, Any]:
        """Navigate back."""
        response = await self._send_command('back', {})
        return response

    @action_logger
    async def forward(self) -> Dict[str, Any]:
        """Navigate forward."""
        response = await self._send_command('forward', {})
        return response

    @action_logger
    async def switch_tab(self, tab_id: str) -> Dict[str, Any]:
        """Switch to a tab."""
        response = await self._send_command('switch_tab', {'tabId': tab_id})
        return response

    @action_logger
    async def close_tab(self, tab_id: str) -> Dict[str, Any]:
        """Close a tab."""
        response = await self._send_command('close_tab', {'tabId': tab_id})
        return response

    @action_logger
    async def get_tab_info(self) -> List[Dict[str, Any]]:
        """Get tab information."""
        response = await self._send_command('get_tab_info', {})
        # The backend returns the tab list directly, not wrapped in an object
        if isinstance(response, list):
            return response
        # Fallback if wrapped in an object
        return response.get('tabs', [])

    @action_logger
    async def console_view(self) -> List[Dict[str, Any]]:
        """Get current page console view"""
        response = await self._send_command('console_view', {})

        if isinstance(response, list):
            return response

        return response.get('logs', [])

    @action_logger
    async def console_exec(self, code: str) -> Dict[str, Any]:
        """Execute javascript code and get result."""
        response = await self._send_command('console_exec', {'code': code})
        return response

    @action_logger
    async def wait_user(
        self, timeout_sec: Optional[float] = None
    ) -> Dict[str, Any]:
        """Wait for user input."""
        response = await self._send_command(
            'wait_user', {'timeout': timeout_sec}
        )
        return response

    async def _read_and_log_output(self):
        """Read stdout from Node.js process & handle SERVER_READY + logging."""
        if not self.process:
            return

        try:
            with contextlib.ExitStack() as stack:
                if self.ts_log_file_path:
                    self.ts_log_file = stack.enter_context(
                        open(self.ts_log_file_path, 'w', encoding='utf-8')
                    )
                    self.ts_log_file.write(
                        f"TypeScript Console Log - Started at "
                        f"{time.strftime('%Y-%m-%d %H:%M:%S')}\n"
                    )
                    self.ts_log_file.write("=" * 80 + "\n")
                    self.ts_log_file.flush()

                while self.process and self.process.poll() is None:
                    try:
                        line = (
                            await asyncio.get_running_loop().run_in_executor(
                                None, self.process.stdout.readline
                            )
                        )
                        if not line:  # EOF
                            break

                        # Check for SERVER_READY message
                        if line.startswith('SERVER_READY:'):
                            try:
                                self.server_port = int(
                                    line.split(':', 1)[1].strip()
                                )
                                logger.info(
                                    f"WebSocket server ready on port "
                                    f"{self.server_port}"
                                )
                                if (
                                    self._server_ready_future
                                    and not self._server_ready_future.done()
                                ):
                                    self._server_ready_future.set_result(True)
                            except (ValueError, IndexError) as e:
                                logger.error(
                                    f"Failed to parse SERVER_READY: {e}"
                                )

                        # Write all output to log file
                        if self.ts_log_file:
                            timestamp = time.strftime('%H:%M:%S')
                            self.ts_log_file.write(f"[{timestamp}] {line}")
                            self.ts_log_file.flush()

                    except Exception as e:
                        logger.warning(f"Error reading stdout: {e}")
                        break

                # Footer if we had a file
                if self.ts_log_file:
                    self.ts_log_file.write("\n" + "=" * 80 + "\n")
                    self.ts_log_file.write(
                        f"TypeScript Console Log - Ended at "
                        f"{time.strftime('%Y-%m-%d %H:%M:%S')}\n"
                    )
                # ExitStack closes file; clear handle
                self.ts_log_file = None
        except Exception as e:
            logger.warning(f"Error in _read_and_log_output: {e}")



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========

# Enables postponed evaluation of annotations (for string-based type hints)
from __future__ import annotations

import base64
import io
import os
from typing import Any, Optional, Union

import requests
from PIL import Image

from camel.embeddings import BaseEmbedding
from camel.types.enums import EmbeddingModelType
from camel.utils import api_keys_required


class JinaEmbedding(BaseEmbedding[Union[str, Image.Image]]):
    r"""Provides text and image embedding functionalities using Jina AI's API.

    Args:
        model_type (EmbeddingModelType, optional): The model to use for
            embeddings. (default: :obj:`JINA_EMBEDDINGS_V3`)
        api_key (Optional[str], optional): The API key for authenticating with
            Jina AI. (default: :obj:`None`)
        dimensions (Optional[int], optional): The dimension of the output
            embeddings. (default: :obj:`None`)
        embedding_type (Optional[str], optional): The type of embedding format
            to generate. Options: 'int8' (binary encoding with higher storage
            and transfer efficiency), 'uint8' (unsigned binary encoding with
            higher storage and transfer efficiency), 'base64' (base64 string
            encoding with higher transfer efficiency). (default: :obj:`None`)
        task (Optional[str], optional): The type of task for text embeddings.
            Options: retrieval.query, retrieval.passage, text-matching,
            classification, separation. (default: :obj:`None`)
        late_chunking (bool, optional): If true, concatenates all sentences in
            input and treats as a single input. (default: :obj:`False`)
        normalized (bool, optional): If true, embeddings are normalized to unit
            L2 norm. (default: :obj:`False`)
    """

    @api_keys_required([("api_key", 'JINA_API_KEY')])
    def __init__(
        self,
        model_type: EmbeddingModelType = EmbeddingModelType.JINA_EMBEDDINGS_V3,
        api_key: Optional[str] = None,
        dimensions: Optional[int] = None,
        embedding_type: Optional[str] = None,
        task: Optional[str] = None,
        late_chunking: bool = False,
        normalized: bool = False,
    ) -> None:
        if not model_type.is_jina:
            raise ValueError(
                f"Model type {model_type} is not a Jina model. "
                "Please use a valid Jina model type."
            )
        self.model_type = model_type
        if dimensions is None:
            self.output_dim = model_type.output_dim
        else:
            self.output_dim = dimensions
        self._api_key = api_key or os.environ.get("JINA_API_KEY")

        self.embedding_type = embedding_type
        self.task = task
        self.late_chunking = late_chunking
        self.normalized = normalized
        self.url = 'https://api.jina.ai/v1/embeddings'
        self.headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json',
            'Authorization': f'Bearer {self._api_key}',
        }

    def embed_list(
        self,
        objs: list[Union[str, Image.Image]],
        **kwargs: Any,
    ) -> list[list[float]]:
        r"""Generates embeddings for the given texts or images.

        Args:
            objs (list[Union[str, Image.Image]]): The texts or images for which
                to generate the embeddings.
            **kwargs (Any): Extra kwargs passed to the embedding API. Not used
                in this implementation.

        Returns:
            list[list[float]]: A list that represents the generated embedding
                as a list of floating-point numbers.

        Raises:
            ValueError: If the input type is not supported.
            RuntimeError: If the API request fails.
        """

        input_data = []
        for obj in objs:
            if isinstance(obj, str):
                if self.model_type == EmbeddingModelType.JINA_CLIP_V2:
                    input_data.append({"text": obj})
                else:
                    input_data.append(obj)  # type: ignore[arg-type]
            elif (
                obj.__class__.__module__ == "PIL.Image"
                and obj.__class__.__name__ == "Image"
            ):
                if self.model_type != EmbeddingModelType.JINA_CLIP_V2:
                    raise ValueError(
                        f"Model {self.model_type} does not support "
                        "image input. Use JINA_CLIP_V2 for image embeddings."
                    )
                # Convert PIL Image to base64 string
                buffered = io.BytesIO()
                obj.save(buffered, format="PNG")
                img_str = base64.b64encode(buffered.getvalue()).decode()
                input_data.append({"image": img_str})
            else:
                raise ValueError(
                    f"Input type {type(obj)} is not supported. "
                    "Must be either str or PIL.Image."
                )

        data = {
            "model": self.model_type.value,
            "input": input_data,
            "embedding_type": "float",
        }

        if self.embedding_type is not None:
            data["embedding_type"] = self.embedding_type
        if self.task is not None:
            data["task"] = self.task
        if self.late_chunking:
            data["late_chunking"] = self.late_chunking  # type: ignore[assignment]
        if self.normalized:
            data["normalized"] = self.normalized  # type: ignore[assignment]
        try:
            response = requests.post(
                self.url, headers=self.headers, json=data, timeout=180
            )
            response.raise_for_status()
            result = response.json()
            return [data["embedding"] for data in result["data"]]
        except requests.exceptions.RequestException as e:
            raise RuntimeError(f"Failed to get embeddings from Jina AI: {e}")

    def get_output_dim(self) -> int:
        r"""Returns the output dimension of the embeddings.

        Returns:
            int: The dimensionality of the embedding for the current model.
        """
        return self.output_dim



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========

import base64
import hashlib
import json
import os
import random
import traceback
from collections import defaultdict
from multiprocessing.pool import ThreadPool
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, Field

from camel.agents.chat_agent import ChatAgent
from camel.benchmarks.base import BaseBenchmark
from camel.logger import get_logger
from camel.societies.role_playing import RolePlaying
from camel.societies.workforce.workforce import Workforce
from camel.tasks.task import Task

logger = get_logger(__name__)


class Message(BaseModel):
    role: str
    content: str
    variant: Optional[str] = None


MessageList = List[Message]


class QueryResponse(BaseModel):
    r"""A structured query response for benchmark evaluation.

    This class defines the expected format for model responses to benchmark
    questions, including explanation, exact answer, and confidence score.
    """

    explanation: str = Field(
        description="""your explanation for your final answer."""
    )
    exact_answer: str = Field(description="""your succinct, final answer.""")
    confidence: str = Field(
        description=r"""
your confidence score between 0|\%| and 100|\%| for your answer.
"""
    )


class GradingResponse(BaseModel):
    r"""A structured grading response for evaluating model answers.

    This class defines the expected format for grading responses, including
    extracted answer, reasoning about correctness, binary correctness judgment,
    and confidence score extraction.
    """

    extracted_final_answer: str = Field(
        description="""
The final exact answer extracted from the [response].
Put the extracted answer as 'None' if there is no exact, final answer to
extract from the response."""
    )
    reasoning: str = Field(
        description="""
Explain why the extracted_final_answer is correct or incorrect
based on [correct_answer], focusing only on if there are meaningful
differences between [correct_answer] and the extracted_final_answer.
Do not comment on any background to the problem, do not attempt
to solve the problem, do not argue for any answer different
than [correct_answer], focus only on whether the answers match."""
    )
    correct: str = Field(
        description="""Answer 'yes' if extracted_final_answer matches the
[correct_answer] given above, or is within a small margin of error for
numerical problems. Answer 'no' otherwise, i.e. if there if there is any
inconsistency, ambiguity, non-equivalency, or if the extracted answer is
incorrect."""
    )
    confidence: str = Field(
        description=r"""The extracted confidence score between 0|\%|
and 100|\%| from [response]. Put 100 if there is no confidence score available.
"""
    )


class SingleEvalResult(BaseModel):
    r"""Result of evaluating a single benchmark sample.

    This class stores the evaluation results for a single benchmark example,
    including score, HTML representation, conversation history, and metrics.
    """

    score: Optional[float] = None
    html: str
    convo: MessageList
    metrics: Dict[str, float] = Field(default_factory=dict)


class EvalResult(BaseModel):
    r"""Result of running a complete benchmark evaluation.

    This class aggregates results from multiple sample evaluations, storing
    the overall score, detailed metrics, HTML reports, and conversation logs.
    """

    score: Optional[float] = None  # top-line metric
    metrics: Optional[Dict[str, float]] = None  # other metrics
    htmls: List[str]  # strings of valid HTML
    convos: List[MessageList]  # sampled conversations


# Define the message template first
_message_template = """
<div class="message {{ role }}">
    <div class="role">
    {{ role }}
    {% if variant %}<span class="variant">({{ variant }})</span>{% endif %}
    </div>
    <div class="content">
    <pre>{{ content }}</pre>
    </div>
</div>
"""

# TODO: Add necessary prompts when tuning.
QUERY_TEMPLATE = """
{question}

Your response should be in the following format:
Explanation: {{your explanation for your final answer}}
Exact Answer: {{your succinct, final answer}}
Confidence: {{your confidence score between 0% and 100% for your answer}}
""".strip()

SUMMARIZE_TEMPLATE = """
Based on the chat history:
{chat_history}

answer the question:
{query}
"""

FORMAT_JSON_TEMPLATE = """
format content into json:
{content}
"""

GRADER_TEMPLATE = r"""
Judge whether the following [response] to [question] is correct or not
based on the precise and unambiguous [correct_answer] below.

[question]: {question}

[response]: {response}

Your judgement must be in the format and criteria specified below:

extracted_final_answer: The final exact answer extracted from the [response].
Put the extracted answer as 'None' if there is no exact, final answer to
Put the extracted answer as 'None' if there is no exact, final answer to
extract from the response.

[correct_answer]: {correct_answer}

reasoning: Explain why the extracted_final_answer is correct or incorrect
based on [correct_answer], focusing only on if there are meaningful
differences between [correct_answer] and the extracted_final_answer.
Do not comment on any background to the problem, do not attempt
to solve the problem, do not argue for any answer different
reasoning: Explain why the extracted_final_answer is correct or incorrect
based on [correct_answer], focusing only on if there are meaningful
differences between [correct_answer] and the extracted_final_answer.
Do not comment on any background to the problem, do not attempt
to solve the problem, do not argue for any answer different
than [correct_answer], focus only on whether the answers match.

correct: Answer 'yes' if extracted_final_answer matches the
[correct_answer] given above, or is within a small margin of error for
numerical problems. Answer 'no' otherwise, i.e. if there is any
inconsistency, ambiguity, non-equivalency, or if the extracted answer is
correct: Answer 'yes' if extracted_final_answer matches the
[correct_answer] given above, or is within a small margin of error for
numerical problems. Answer 'no' otherwise, i.e. if there is any
inconsistency, ambiguity, non-equivalency, or if the extracted answer is
incorrect.


confidence: The extracted confidence score between 0|\%| and 100|\%|
confidence: The extracted confidence score between 0|\%| and 100|\%|
from [response]. Put 100 if there is no confidence score available.
""".strip()


HTML_JINJA = """
<h3>Question:</h3>
{{ message_to_html(prompt_messages) | safe }}
<h3>Sampled message</h3>
{{ message_to_html(next_message) | safe }}
<h3>Results</h3>
<p>Correct Answer: {{ correct_answer }}</p>
<p>Extracted Answer: {{ extracted_answer }}</p>
<p>Score: {{ score }}</p>
"""
_report_template = """<!DOCTYPE html>
<html>
    <head>
        <style>
            .message {
                padding: 8px 16px;
                margin-bottom: 8px;
                border-radius: 4px;
            }
            .message.user {
                background-color: #B2DFDB;
                color: #00695C;
            }
            .message.assistant {
                background-color: #B39DDB;
                color: #4527A0;
            }
            .message.system {
                background-color: #EEEEEE;
                color: #212121;
            }
            .role {
                font-weight: bold;
                margin-bottom: 4px;
            }
            .variant {
                color: #795548;
            }
            table, th, td {
                border: 1px solid black;
            }
            pre {
                white-space: pre-wrap;
            }
        </style>
    </head>
    <body>
    {% if metrics %}
    <h1>Metrics</h1>
    <table>
    <tr>
        <th>Metric</th>
        <th>Value</th>
    </tr>
    <tr>
        <td><b>Score</b></td>
        <td>{{ score | float | round(3) }}</td>
    </tr>
    {% for name, value in metrics.items() %}
    <tr>
        <td>{{ name }}</td>
        <td>{{ value }}</td>
    </tr>
    {% endfor %}
    </table>
    {% endif %}
    <h1>Examples</h1>
    {% for html in htmls %}
    {{ html | safe }}
    <hr>
    {% endfor %}
    </body>
</html>
"""


class JinjaEnv:
    r"""A class that encapsulates the Jinja environment setup."""

    _instance: Optional['JinjaEnv'] = None
    _env = None

    def __init__(self):
        r"""Initialize the JinjaEnv instance if not already initialized."""
        if not getattr(self, '_initialized', False):
            self._initialized = True

    def __new__(cls):
        r"""Implement singleton pattern to ensure only one instance exists."""
        if cls._instance is None:
            cls._instance = super(JinjaEnv, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    @classmethod
    def get_instance(cls):
        r"""Get the singleton instance of JinjaEnv.

        Returns:
            JinjaEnv: The singleton instance.
        """
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    @property
    def env(self):
        r"""Lazily initialize and return the Jinja environment.

        Returns:
            jinja2.Environment: The Jinja environment instance.
        """
        if self._env is None:
            # Lazy import of jinja2
            import jinja2

            # Create the Jinja environment
            self._env = jinja2.Environment(
                loader=jinja2.BaseLoader(),
                undefined=jinja2.StrictUndefined,
                autoescape=jinja2.select_autoescape(["html", "xml"]),
            )

            # Register the message_to_html function
            self._env.globals["message_to_html"] = self.message_to_html

        return self._env

    def from_string(self, template_str):
        r"""Create a template from the given string.

        Args:
            template_str (str): The template string.

        Returns:
            jinja2.Template: The compiled template.
        """
        return self.env.from_string(template_str)

    @staticmethod
    def message_to_html(message: Message) -> str:
        r"""Generate HTML snippet (inside a <div>) for a message.

        Args:
            message (Message): The message to convert to HTML.

        Returns:
            str: The HTML representation of the message.
        """
        return (
            JinjaEnv.get_instance()
            .from_string(_message_template)
            .render(
                role=message.role,
                content=message.content,
                variant=message.variant,
            )
        )


def derive_key(password: str, length: int) -> bytes:
    r"""Derive a fixed-length key from the password using SHA256."""
    hasher = hashlib.sha256()
    hasher.update(password.encode())
    key = hasher.digest()
    return key * (length // len(key)) + key[: length % len(key)]


def decrypt(ciphertext_b64: str, password: str) -> str:
    r"""Decrypt base64-encoded ciphertext with XOR."""
    encrypted = base64.b64decode(ciphertext_b64)
    key = derive_key(password, len(encrypted))
    decrypted = bytes(a ^ b for a, b in zip(encrypted, key))
    return decrypted.decode()


def _compute_stat(values: list, stat: str):
    import numpy as np

    if stat == "mean":
        return np.mean(values)
    elif stat == "std":
        return np.std(values)
    elif stat == "min":
        return np.min(values)
    elif stat == "max":
        return np.max(values)
    else:
        raise ValueError(f"Unknown {stat =}")


def aggregate_results(
    single_eval_results: List[SingleEvalResult],
    default_stats: Tuple[str, str] = ("mean", "std"),
    name2stats: Optional[Dict[str, Tuple[str]]] = None,
) -> EvalResult:
    r"""Aggregate results from multiple evaluations into a single EvalResult.

    Args:
        single_eval_results (List[SingleEvalResult]): A list of
            `SingleEvalResult` objects.
        default_stats (Tuple[str, str]): A tuple of default statistics to
            compute. (default: :obj:`("mean", "std")`)
        name2stats (Optional[Dict[str, Tuple[str]]]): A dictionary mapping
            metric names to statistics to compute. (default: :obj:`None`)

    Returns:
        EvalResult: An `EvalResult` object containing aggregated results.
    """
    name2stats = name2stats or {}
    name2values = defaultdict(list)
    htmls = []
    convos = []

    for single_eval_result in single_eval_results:
        for name, value in single_eval_result.metrics.items():
            name2values[name].append(value)
        if single_eval_result.score is not None:
            name2values["score"].append(single_eval_result.score)
        htmls.append(single_eval_result.html)
        convos.append(single_eval_result.convo)

    final_metrics = {}
    for name, values in name2values.items():
        stats = name2stats.get(name, default_stats)
        for stat in stats:
            key = name if stat == "mean" else f"{name}:{stat}"
            final_metrics[key] = _compute_stat(values, stat)

    return EvalResult(
        score=final_metrics.pop("score", None),
        metrics=final_metrics,
        htmls=htmls,
        convos=convos,
    )


class BrowseCompBenchmark(BaseBenchmark):
    r"""BrowseComp Benchmark for evaluating browser-based comprehension tasks.

    This benchmark evaluates the ability of language models to comprehend and
    answer questions based on browser-based content, measuring accuracy and
    performance.
    """

    def __init__(
        self,
        save_to: str,
        processes: int = 1,
        num_examples: Optional[int] = None,
        n_repeats: int = 1,
    ):
        r"""Initialize the BrowseComp benchmark.

        Args:
            save_to (str): The file to save the results.
            processes (int, optional): The number of processes to use for
                parallel processing. (default: :obj:`1`)
            num_examples (Optional[int]): Number of examples to evaluate.
                If None, all examples are used. Controls the sample size for
                testing. (default: :obj:`None`)
            n_repeats (int, optional): Number of times to repeat each example.
                Useful for evaluating consistency across multiple runs.
                (default: :obj:`1`)
        """
        # Browsecomp benchmark won't download any data
        # use current path as the data_dir passing into super init
        current_path = os.path.dirname(os.path.abspath(__file__))

        super().__init__("browsecomp", current_path, save_to, processes)
        self.num_examples = num_examples
        self.n_repeats = n_repeats
        self.examples: List[Dict[str, Any]] = []
        self.load()
        self._raw_results: List[Any] = []
        self._validated_results: List[SingleEvalResult] = []
        self._eval_result: EvalResult
        self.jinja_env = JinjaEnv.get_instance()

    def download(self):
        r"""Download the BrowseComp dataset.

        This method is implemented to maintain compatibility
        with the BaseBenchmark interface, but BrowseComp doesn't
        require downloading data separately.

        Returns:
            self: The benchmark instance
        """
        logger.info("BrowseComp benchmark does not require downloading data.")
        return self

    def load(self):
        r"""Load the BrowseComp dataset.

        This method loads the dataset from a remote CSV file, converts each
        row to a dictionary, and applies sampling if num_examples is
        specified. It also handles repeating examples if n_repeats > 1.

        Returns:
            self: The benchmark instance
        """
        # Load dataset from remote CSV
        import pandas

        df = pandas.read_csv(
            "https://openaipublic.blob.core.windows.net/simple-evals/browse_comp_test_set.csv"
        )
        # Convert each row to a dictionary
        examples = [row.to_dict() for _, row in df.iterrows()]

        # Sample examples if num_examples is specified
        if self.num_examples:
            assert (
                self.n_repeats == 1
            ), "n_repeats only supported when max_examples = None"
            rng = random.Random(0)  # Use fixed seed for reproducibility
            examples = rng.sample(examples, self.num_examples)

        # Repeat examples if n_repeats > 1
        self.examples = examples * self.n_repeats
        return self

    @property
    def train(self):
        r"""Get the training set.

        This property is implemented to maintain compatibility with
        the BaseBenchmark interface, but BrowseComp doesn't have a
        training set.

        Raises:
            NotImplementedError: BrowseComp does not have a training set.
        """
        raise NotImplementedError("BrowseComp does not have a training set.")

    def run(  # type: ignore[override]
        self,
        pipeline_template: Union[ChatAgent, RolePlaying, Workforce],
        chat_turn_limit: int = 10,
        roleplaying_summarizer: Optional[ChatAgent] = None,
        task_json_formatter: Optional[ChatAgent] = None,
    ) -> None:
        r"""Run the benchmark by processing each example in parallel.

        This method applies the provided pipeline to each example in the
        dataset using a process pool for parallel execution. It shows progress
        using tqdm and stores the results in self._raw_results.

        Args:
            pipeline_template (Union[ChatAgent, RolePlaying, Workforce]): The
                template agent or framework to use for processing examples.
                Can be a ChatAgent, RolePlaying, or Workforce instance that
                will be cloned for each example.
            chat_turn_limit (int): Maximum number of conversation turns allowed
                when using RolePlaying pipeline. (default: :obj:`10`)
            roleplaying_summarizer (Optional[ChatAgent]): Optional ChatAgent to
                summarize RolePlaying conversations. If None and RolePlaying is
                used, a default summarizer will be created.
                (default: :obj:`None`)
            task_json_formatter (Optional[ChatAgent]): Optional ChatAgent to
                format task JSON. If None and Workforce is used, a default
                formatter will be created. (default: :obj:`None`)
        """
        from tqdm import tqdm

        # Use a process pool for parallel execution
        def process_benchmark_row(row: Dict[str, Any]) -> Dict[str, Any]:
            r"""This inner function processes a single benchmark row by
            extracting the problem and answer, creating a pipeline instance,
            and generating a response using the appropriate method based on
            the pipeline type.

            Args:
                row (Dict[str, Any]): A row from the dataset containing
                    encrypted problem and answer, along with a canary for
                    decryption.

            Returns:
                Dict[str, Any]: A dictionary containing the decrypted problem,
                    expected answer, model response, and structured response
                    fields.
            """

            problem = decrypt(row.get("problem", ""), row.get("canary", ""))
            answer = decrypt(row.get("answer", ""), row.get("canary", ""))
            try:
                input_message = QUERY_TEMPLATE.format(question=problem)

                if isinstance(pipeline_template, (ChatAgent)):
                    pipeline = pipeline_template.clone()  # type: ignore[assignment]

                    response_text = pipeline.step(
                        input_message, response_format=QueryResponse
                    )
                elif isinstance(pipeline_template, Workforce):
                    pipeline = pipeline_template.clone()  # type: ignore[assignment]
                    task = Task(content=input_message, id="0")
                    task = pipeline.process_task(task)  # type: ignore[attr-defined]
                    if task_json_formatter:
                        formatter_in_process = task_json_formatter.clone()
                    else:
                        formatter_in_process = ChatAgent(
                            "You are a helpful assistant."
                        )
                    response_text = formatter_in_process.step(
                        FORMAT_JSON_TEMPLATE.format(content=task.result),
                        response_format=QueryResponse,
                    )

                elif isinstance(pipeline_template, RolePlaying):
                    # RolePlaying is different.
                    pipeline = pipeline_template.clone(  # type: ignore[assignment]
                        task_prompt=input_message
                    )

                    n = 0
                    input_msg = pipeline.init_chat()  # type: ignore[attr-defined]
                    chat_history = []
                    while n < chat_turn_limit:
                        n += 1
                        assistant_response, user_response = pipeline.step(
                            input_msg
                        )
                        if assistant_response.terminated:  # type: ignore[union-attr]
                            break
                        if user_response.terminated:  # type: ignore[union-attr]
                            break
                        if "CAMEL_TASK_DONE" in user_response.msg.content:  # type: ignore[union-attr]
                            break

                        chat_history.append(
                            f"AI User: {user_response.msg.content}"  # type: ignore[union-attr]
                        )
                        chat_history.append(
                            f"AI Assistant: {assistant_response.msg.content}"  # type: ignore[union-attr]
                        )
                        input_msg = assistant_response.msg  # type: ignore[union-attr]

                    chat_history_str = "\n".join(chat_history)
                    if roleplaying_summarizer:
                        summarizer_in_process = roleplaying_summarizer.clone()
                    else:
                        summarizer_in_process = ChatAgent(
                            "You are a helpful assistant."
                        )

                    summarize_prompt = SUMMARIZE_TEMPLATE.format(
                        chat_history=chat_history_str,
                        query=input_message,
                    )
                    response_text = summarizer_in_process.step(
                        summarize_prompt, response_format=QueryResponse
                    )
                else:
                    raise NotImplementedError(
                        f"{type(pipeline_template)} is not supported."
                    )
                # Parse the response JSON
                response_dict = json.loads(response_text.msg.content)

                # Format the response as a key-value string
                formatted_response = f"""
    Explanation: {response_dict['explanation']}

    Exact Answer: {response_dict['exact_answer']}
    Confidence: {response_dict['confidence']}"""

                # Create the result dictionary
                raw_result = {}
                raw_result['problem'] = problem
                raw_result['expected_answer'] = answer
                raw_result['response'] = formatted_response
                # Keep the original dict for reference
                raw_result['response_dict'] = response_dict

                return raw_result
            except Exception as e:
                # Log any errors that occur during evaluation
                logger.error(f"Error evaluating result: {e}")
                logger.error(traceback.format_exc())
                return {
                    'problem': problem,
                    'expected_answer': answer,
                    'response': traceback.format_exc(),
                    'response_dict': {},
                }

        pool_class = ThreadPool
        with pool_class(min(self.processes, len(self.examples))) as pool:
            self._raw_results = list(
                tqdm(
                    pool.imap(process_benchmark_row, self.examples),
                    total=len(self.examples),
                )
            )

    def make_report(self, eval_result: EvalResult) -> str:
        r"""Create a standalone HTML report from an EvalResult."""
        return self.jinja_env.from_string(_report_template).render(
            score=eval_result.score,
            metrics=eval_result.metrics,
            htmls=eval_result.htmls,
        )

    def validate(self, grader: Optional[ChatAgent] = None) -> None:
        r"""Validate the raw results using the GRADER_TEMPLATE and ChatAgent.

        This method evaluates the correctness of each response by
        multi-threading. A dedicated chat agent is created in each thread.
        The chat agent will compare raw result with the expected answer. The
        grading results will be aggregated in a report.

        Args:
            grader: The ChatAgent used for validation. If None, a default
                agent will be created in each thread. If provided, the
                provided agent will be used as a template and be cloned into
                new agents in each thread. (default: :obj:`None`)
        """
        from tqdm import tqdm

        def validate_each_one(raw_result: Dict[str, Any]) -> SingleEvalResult:
            r"""This inner function formats the prompt for the ChatAgent
            grader, sends it for evaluation, extracts the correctness
            assessment, and creates an HTML representation of the result.

            Args:
                raw_result (Dict[str, Any]): A dictionary containing 'problem',
                    'response', and 'expected_answer' keys.

            Returns:
                SingleEvalResult: An evaluation result object with score,
                    metrics, and HTML.
            """
            # Format the template
            prompt = GRADER_TEMPLATE.format(
                question=raw_result['problem'],
                response=raw_result['response'],
                correct_answer=raw_result['expected_answer'],
            )
            if grader:
                grader_in_process = grader.clone()
            else:
                grader_in_process = ChatAgent("You are a helpful assistant.")

            # Create a conversation list for the result
            convo = [
                Message(content=raw_result['problem'], role="user"),
                Message(content=raw_result['response'], role="assistant"),
            ]

            try:
                response = grader_in_process.step(
                    prompt, response_format=GradingResponse
                )

                content = json.loads(response.msg.content)

                grade_result = content['correct']

                # Convert to binary metrics (1 for correct, 0 for incorrect)
                is_correct = int(grade_result == "yes")
                is_incorrect = int(grade_result == "no")

                # Set the score (1 for correct, 0 for incorrect)
                score = is_correct

                # Generate HTML representation of the result
                html = self.jinja_env.from_string(HTML_JINJA).render(
                    prompt_messages=Message(
                        content=raw_result.get('problem', ''), role="user"
                    ),
                    next_message=Message(
                        content=raw_result.get('response', ''),
                        role="assistant",
                    ),
                    score=score,
                    correct_answer=raw_result.get('expected_answer', ''),
                    extracted_answer=raw_result.get('response_dict', {}).get(
                        'exact_answer', ''
                    ),
                )
                # Return the evaluation result
                return SingleEvalResult(
                    html=html,
                    score=score,
                    convo=convo,
                    metrics={
                        "is_correct": is_correct,
                        "is_incorrect": is_incorrect,
                    },
                )
            except Exception as e:
                # Log any errors that occur during evaluation
                logger.error(f"Error evaluating result: {e}")
                logger.error(traceback.format_exc())
                html = self.jinja_env.from_string(HTML_JINJA).render(
                    prompt_messages=Message(
                        content=raw_result.get('problem', ''), role="user"
                    ),
                    next_message=Message(
                        content=raw_result.get('response', ''),
                        role="assistant",
                    ),
                    score=0,
                    correct_answer=raw_result.get('expected_answer', ''),
                    extracted_answer=raw_result.get('response_dict', {}).get(
                        'exact_answer', ''
                    ),
                )
                return SingleEvalResult(
                    html=html,
                    score=0,
                    convo=convo,
                    metrics={
                        "is_correct": 0,
                        "is_incorrect": 1,
                    },
                )

        pool_class = ThreadPool
        with pool_class(min(self.processes, len(self._raw_results))) as pool:
            self._validated_results = list(
                tqdm(
                    pool.imap(validate_each_one, self._raw_results),
                    total=len(self._raw_results),
                )
            )

        aggregate_metrics = {
            "is_correct": sum(
                result.metrics["is_correct"]
                for result in self._validated_results
            )
            / len(self._validated_results),
            "is_incorrect": sum(
                result.metrics["is_incorrect"]
                for result in self._validated_results
            )
            / len(self._validated_results),
        }
        logger.info("AGGREGATE METRICS")
        logger.info(aggregate_metrics)
        logger.info("##################")

        output_d = {
            "accuracy": aggregate_metrics["is_correct"],
        }

        logger.info(f"Accuracy: {output_d['accuracy']:.3f}")

        self._eval_result = aggregate_results(self._validated_results)
        # ^^^ how to use a sampler
        report_filename = self.save_to
        logger.info(f"Writing report to {report_filename}")
        with open(report_filename, "w") as fh:
            fh.write(self.make_report(self._eval_result))



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
from collections import deque
from typing import Dict, List, Optional

from camel.agents import (
    ChatAgent,
    TaskCreationAgent,
    TaskPrioritizationAgent,
    TaskSpecifyAgent,
)
from camel.agents.chat_agent import ChatAgentResponse
from camel.generators import SystemMessageGenerator
from camel.logger import get_logger
from camel.messages import BaseMessage
from camel.prompts import TextPrompt
from camel.types import RoleType, TaskType

logger = get_logger(__name__)


class BabyAGI:
    r"""The BabyAGI Agent adapted from `"Task-driven Autonomous Agent"
    <https://github.com/yoheinakajima/babyagi>`_.

    Args:
        assistant_role_name (str): The name of the role played by the
            assistant.
        user_role_name (str): The name of the role played by the user.
        task_prompt (str, optional): A prompt for the task to be performed.
            (default: :obj:`""`)
        task_type (TaskType, optional): The type of task to perform.
            (default: :obj:`TaskType.AI_SOCIETY`)
        max_task_history (int): The maximum number of previous tasks
            information to include in the task agent.
            (default: :obj:10)
        assistant_agent_kwargs (Dict, optional): Additional arguments to pass
            to the assistant agent. (default: :obj:`None`)
        task_specify_agent_kwargs (Dict, optional): Additional arguments to
            pass to the task specify agent. (default: :obj:`None`)
        task_creation_agent_kwargs (Dict, optional): Additional arguments to
            pass to the task creation agent. (default: :obj:`None`)
        task_prioritization_agent_kwargs (Dict, optional): Additional arguments
            to pass to the task prioritization agent. (default: :obj:`None`)
        sys_msg_generator_kwargs (Dict, optional): Additional arguments to
            pass to the system message generator. (default: :obj:`None`)
        extend_task_specify_meta_dict (Dict, optional): A dict to extend the
            task specify meta dict with. (default: :obj:`None`)
        output_language (str, optional): The language to be output by the
            agents. (default: :obj:`None`)
        message_window_size (int, optional): The maximum number of previous
            messages to include in the context window. If `None`, no windowing
            is performed. (default: :obj:`None`)
    """

    def __init__(
        self,
        assistant_role_name: str,
        user_role_name: str,
        task_prompt: str = "",
        task_type: TaskType = TaskType.AI_SOCIETY,
        max_task_history: int = 10,
        assistant_agent_kwargs: Optional[Dict] = None,
        task_specify_agent_kwargs: Optional[Dict] = None,
        task_creation_agent_kwargs: Optional[Dict] = None,
        task_prioritization_agent_kwargs: Optional[Dict] = None,
        sys_msg_generator_kwargs: Optional[Dict] = None,
        extend_task_specify_meta_dict: Optional[Dict] = None,
        output_language: Optional[str] = None,
        message_window_size: Optional[int] = None,
    ) -> None:
        self.task_type = task_type
        self.task_prompt = task_prompt
        self.specified_task_prompt: TextPrompt
        self.init_specified_task_prompt(
            assistant_role_name,
            user_role_name,
            task_specify_agent_kwargs,
            extend_task_specify_meta_dict,
            output_language,
        )

        sys_msg_generator = SystemMessageGenerator(
            task_type=self.task_type, **(sys_msg_generator_kwargs or {})
        )

        init_assistant_sys_msg = sys_msg_generator.from_dicts(
            meta_dicts=[
                dict(
                    assistant_role=assistant_role_name,
                    user_role=user_role_name,
                    task=self.specified_task_prompt,
                )
            ],
            role_tuples=[
                (assistant_role_name, RoleType.ASSISTANT),
            ],
        )

        self.assistant_agent: ChatAgent
        self.assistant_sys_msg: Optional[BaseMessage]
        self.task_creation_agent: TaskCreationAgent
        self.task_prioritization_agent: TaskPrioritizationAgent
        self.init_agents(
            init_assistant_sys_msg[0],
            assistant_agent_kwargs,
            task_creation_agent_kwargs,
            task_prioritization_agent_kwargs,
            output_language,
            message_window_size,
        )

        self.subtasks: deque = deque([])
        self.solved_subtasks: List[str] = []
        self.MAX_TASK_HISTORY = max_task_history

    def init_specified_task_prompt(
        self,
        assistant_role_name: str,
        user_role_name: str,
        task_specify_agent_kwargs: Optional[Dict],
        extend_task_specify_meta_dict: Optional[Dict],
        output_language: Optional[str],
    ):
        r"""Use a task specify agent to generate a specified task prompt.
        Generated specified task prompt will be used to replace original
        task prompt. If there is no task specify agent, specified task
        prompt will not be generated.

        Args:
            assistant_role_name (str): The name of the role played by the
                assistant.
            user_role_name (str): The name of the role played by the user.
            task_specify_agent_kwargs (Dict, optional): Additional arguments
                to pass to the task specify agent.
            extend_task_specify_meta_dict (Dict, optional): A dict to extend
                the task specify meta dict with.
            output_language (str, optional): The language to be output by the
                agents.
        """
        task_specify_meta_dict = dict()
        if self.task_type in [TaskType.AI_SOCIETY, TaskType.MISALIGNMENT]:
            task_specify_meta_dict.update(
                dict(
                    assistant_role=assistant_role_name,
                    user_role=user_role_name,
                )
            )
        task_specify_meta_dict.update(extend_task_specify_meta_dict or {})
        task_specify_agent = TaskSpecifyAgent(
            task_type=self.task_type,
            output_language=output_language,
            **(task_specify_agent_kwargs or {}),
        )
        self.specified_task_prompt = task_specify_agent.run(
            self.task_prompt,
            meta_dict=task_specify_meta_dict,
        )

    def init_agents(
        self,
        init_assistant_sys_msg: BaseMessage,
        assistant_agent_kwargs: Optional[Dict],
        task_creation_agent_kwargs: Optional[Dict],
        task_prioritization_agent_kwargs: Optional[Dict],
        output_language: Optional[str],
        message_window_size: Optional[int] = None,
    ):
        r"""Initialize assistant and user agents with their system messages.

        Args:
            init_assistant_sys_msg (BaseMessage): Assistant agent's initial
                system message.
            assistant_agent_kwargs (Dict, optional): Additional arguments to
                pass to the assistant agent.
            task_creation_agent_kwargs (Dict, optional): Additional arguments
                to pass to the task creation agent.
            task_prioritization_agent_kwargs (Dict, optional): Additional
                arguments to pass to the task prioritization agent.
            output_language (str, optional): The language to be output by the
                agents.
            message_window_size (int, optional): The maximum number of previous
                messages to include in the context window. If `None`, no
                windowing is performed. (default: :obj:`None`)
        """
        self.assistant_agent = ChatAgent(
            init_assistant_sys_msg,
            output_language=output_language,
            message_window_size=message_window_size,
            **(assistant_agent_kwargs or {}),
        )
        self.assistant_sys_msg = self.assistant_agent.system_message
        self.assistant_agent.reset()

        self.task_creation_agent = TaskCreationAgent(
            objective=self.specified_task_prompt,
            role_name=getattr(self.assistant_sys_msg, 'role_name', None)
            or "assistant",
            output_language=output_language,
            message_window_size=message_window_size,
            **(task_creation_agent_kwargs or {}),
        )
        self.task_creation_agent.reset()

        self.task_prioritization_agent = TaskPrioritizationAgent(
            objective=self.specified_task_prompt,
            output_language=output_language,
            message_window_size=message_window_size,
            **(task_prioritization_agent_kwargs or {}),
        )
        self.task_prioritization_agent.reset()

    def step(self) -> ChatAgentResponse:
        r"""BabyAGI agent would pull the first task from the task list,
        complete the task based on the context, then creates new tasks and
        re-prioritizes the task list based on the objective and the result of
        the previous task. It returns assistant message.

        Returns:
            ChatAgentResponse: it contains the resulting assistant message,
            whether the assistant agent terminated the conversation,
            and any additional assistant information.

        """
        if not self.subtasks:
            new_subtask_list = self.task_creation_agent.run(task_list=[])
            prioritized_subtask_list = self.task_prioritization_agent.run(
                new_subtask_list
            )
            self.subtasks = deque(prioritized_subtask_list)

        task_name = self.subtasks.popleft()
        assistant_msg_msg = BaseMessage.make_user_message(
            role_name=getattr(self.assistant_sys_msg, 'role_name', None)
            or "assistant",
            content=f"{task_name}",
        )

        assistant_response = self.assistant_agent.step(assistant_msg_msg)
        assistant_msg = assistant_response.msgs[0]

        self.solved_subtasks.append(task_name)
        past_tasks = self.solved_subtasks + list(self.subtasks)

        new_subtask_list = self.task_creation_agent.run(
            task_list=past_tasks[-self.MAX_TASK_HISTORY :]
        )

        if new_subtask_list:
            self.subtasks.extend(new_subtask_list)
            prioritized_subtask_list = self.task_prioritization_agent.run(
                task_list=list(self.subtasks)[-self.MAX_TASK_HISTORY :]
            )
            self.subtasks = deque(prioritized_subtask_list)
        else:
            logger.info("no new tasks")
        assistant_response.info['task_name'] = task_name
        assistant_response.info['subtasks'] = list(self.subtasks)
        if not self.subtasks:
            terminated = True
            assistant_response.info['termination_reasons'] = (
                "All tasks are solved"
            )
            return ChatAgentResponse(
                msgs=[assistant_msg],
                terminated=terminated,
                info=assistant_response.info,
            )
        return ChatAgentResponse(
            msgs=[assistant_msg],
            terminated=assistant_response.terminated,
            info=assistant_response.info,
        )



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
from __future__ import annotations

import asyncio
import logging
from abc import ABC, abstractmethod
from typing import List, Optional, Set

from colorama import Fore

from camel.societies.workforce.base import BaseNode
from camel.societies.workforce.task_channel import TaskChannel
from camel.societies.workforce.utils import check_if_running
from camel.tasks.task import Task, TaskState

logger = logging.getLogger(__name__)


class Worker(BaseNode, ABC):
    r"""A worker node that works on tasks. It is the basic unit of task
    processing in the workforce system.

    Args:
        description (str): Description of the node.
        node_id (Optional[str]): ID of the node. If not provided, it will
            be generated automatically. (default: :obj:`None`)
    """

    def __init__(
        self,
        description: str,
        node_id: Optional[str] = None,
    ) -> None:
        super().__init__(description, node_id=node_id)
        self._active_task_ids: Set[str] = set()
        self._running_tasks: Set[asyncio.Task] = set()

    def __repr__(self):
        return f"Worker node {self.node_id} ({self.description})"

    @abstractmethod
    async def _process_task(
        self, task: Task, dependencies: List[Task]
    ) -> TaskState:
        r"""Processes a task based on its dependencies.

        Returns:
            'DONE' if the task is successfully processed,
            'FAILED' if the processing fails.
        """
        pass

    async def _get_assigned_task(self) -> Task:
        r"""Get a task assigned to this node from the channel."""
        return await self._channel.get_assigned_task_by_assignee(self.node_id)

    @staticmethod
    def _get_dep_tasks_info(dependencies: List[Task]) -> str:
        result_lines = [
            f"id: {dep_task.id}, content: {dep_task.content}. "
            f"result: {dep_task.result}."
            for dep_task in dependencies
        ]
        result_str = "\n".join(result_lines)
        return result_str

    @check_if_running(False)
    def set_channel(self, channel: TaskChannel):
        self._channel = channel

    async def _process_single_task(self, task: Task) -> None:
        r"""Process a single task and handle its completion/failure."""
        try:
            self._active_task_ids.add(task.id)
            print(
                f"{Fore.YELLOW}{self} get task {task.id}: {task.content}"
                f"{Fore.RESET}"
            )

            # Process the task
            task_state = await self._process_task(task, task.dependencies)

            # Update the result and status of the task
            task.set_state(task_state)

            await self._channel.return_task(task.id)
        except Exception as e:
            logger.error(f"Error processing task {task.id}: {e}")
            # Store error information in task result
            task.result = f"{type(e).__name__}: {e!s}"
            task.set_state(TaskState.FAILED)
            await self._channel.return_task(task.id)
        finally:
            self._active_task_ids.discard(task.id)

    @check_if_running(False)
    async def _listen_to_channel(self):
        r"""Continuously listen to the channel and process assigned tasks.

        This method supports parallel task execution without artificial limits.
        """
        self._running = True
        logger.info(f"{self} started.")

        while self._running:
            try:
                # Clean up completed tasks
                completed_tasks = [t for t in self._running_tasks if t.done()]
                for completed_task in completed_tasks:
                    self._running_tasks.discard(completed_task)
                    # Check for exceptions in completed tasks
                    try:
                        await completed_task
                    except Exception as e:
                        logger.error(f"Task processing failed: {e}")

                # Try to get a new task (with short timeout to avoid blocking)
                try:
                    task = await asyncio.wait_for(
                        self._get_assigned_task(), timeout=1.0
                    )

                    # Create and start processing task
                    task_coroutine = asyncio.create_task(
                        self._process_single_task(task)
                    )
                    self._running_tasks.add(task_coroutine)

                except asyncio.TimeoutError:
                    # No tasks available, continue loop
                    if not self._running_tasks:
                        # No tasks running and none available, short sleep
                        await asyncio.sleep(0.1)
                    continue

            except Exception as e:
                logger.error(
                    f"Error in worker {self.node_id} listen loop: {e}"
                )
                await asyncio.sleep(0.1)
                continue

        # Wait for all remaining tasks to complete when stopping
        if self._running_tasks:
            logger.info(
                f"{self} stopping, waiting for {len(self._running_tasks)} "
                f"tasks to complete..."
            )
            await asyncio.gather(*self._running_tasks, return_exceptions=True)

        logger.info(f"{self} stopped.")

    @check_if_running(False)
    async def start(self):
        r"""Start the worker."""
        await self._listen_to_channel()

    @check_if_running(True)
    def stop(self):
        r"""Forcefully stop the worker.

        Cancels all running tasks immediately and sets the stop flag.
        The worker will exit after completing cancellation.
        """
        # First cancel all running tasks to interrupt ongoing work
        tasks_to_cancel = list(self._running_tasks)
        for task in tasks_to_cancel:
            if not task.done():
                task.cancel()

        # Clear the running tasks set since they're all cancelled
        self._running_tasks.clear()

        # Set stop flag to exit the listen loop
        self._running = False
        return



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
import asyncio
import concurrent.futures
import importlib
import io
import json
import logging
import os
import sys
from typing import Any, Dict, List

import uvicorn
from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse

from camel.toolkits import BaseToolkit

# thread pool for running sync tools that can't run inside async event loop
# (e.g., Playwright sync API)
_executor = concurrent.futures.ThreadPoolExecutor(max_workers=4)

logger = logging.getLogger(__name__)

# set environment variable to indicate we're running inside a CAMEL runtime
os.environ["CAMEL_RUNTIME"] = "true"

sys.path.append(os.getcwd())

modules_functions = sys.argv[1:]

logger.info(f"Modules and functions: {modules_functions}")

app = FastAPI()

# global cache for toolkit instances to maintain state across calls
_toolkit_instances: Dict[str, Any] = {}

# track registered endpoints for health check
_registered_endpoints: List[str] = []


@app.get("/health")
async def health_check():
    r"""Health check endpoint that reports loaded toolkits and endpoints."""
    return {
        "status": "ok",
        "toolkits": list(_toolkit_instances.keys()),
        "endpoints": _registered_endpoints,
    }


@app.exception_handler(Exception)
async def general_exception_handler(request: Request, exc: Exception):
    return JSONResponse(
        status_code=500,
        content={
            "detail": "Internal Server Error",
            "error_message": str(exc),
        },
    )


for module_function in modules_functions:
    try:
        # store original module_function as cache key before parsing
        cache_key = module_function

        init_params = dict()
        if "{" in module_function:
            module_function, params = module_function.split("{")
            params = "{" + params
            init_params = json.loads(params)

        module_name, function_name = module_function.rsplit(".", 1)

        logger.info(f"Importing {module_name} and function {function_name}")

        module = importlib.import_module(module_name)
        function = getattr(module, function_name)
        if isinstance(function, type) and issubclass(function, BaseToolkit):
            # use cached instance if available to maintain state across calls
            if cache_key not in _toolkit_instances:
                _toolkit_instances[cache_key] = function(**init_params)
            function = _toolkit_instances[cache_key].get_tools()

        if not isinstance(function, list):
            function = [function]

        for func in function:
            endpoint_name = func.get_function_name()
            _registered_endpoints.append(endpoint_name)

            def make_endpoint(tool):
                r"""Create endpoint with tool captured in closure."""

                def run_tool(data: Dict):
                    r"""Run tool in thread pool to avoid async event loop."""
                    redirect_stdout = data.get('redirect_stdout', False)
                    captured_output = None
                    if redirect_stdout:
                        captured_output = io.StringIO()
                        old_stdout = sys.stdout
                        sys.stdout = captured_output
                    try:
                        response_data = tool.func(
                            *data['args'], **data['kwargs']
                        )
                    finally:
                        if redirect_stdout:
                            sys.stdout = old_stdout
                    if redirect_stdout and captured_output is not None:
                        captured_output.seek(0)
                        output = captured_output.read()
                        return {
                            "output": json.dumps(
                                response_data, ensure_ascii=False
                            ),
                            "stdout": output,
                        }
                    return {
                        "output": json.dumps(response_data, ensure_ascii=False)
                    }

                async def endpoint(data: Dict):
                    # run in thread pool to support sync tools like Playwright
                    loop = asyncio.get_running_loop()
                    return await loop.run_in_executor(
                        _executor, run_tool, data
                    )

                return endpoint

            app.post(f"/{endpoint_name}")(make_endpoint(func))

    except (ImportError, AttributeError) as e:
        logger.error(f"Error importing {module_function}: {e}")


if __name__ == "__main__":
    # reload=False to avoid conflicts with async toolkits (e.g., Playwright)
    uvicorn.run(app, host="0.0.0.0", port=8000, reload=False)



# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ========= Copyright 2023-2026 @ CAMEL-AI.org. All Rights Reserved. =========

import uuid
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Literal, Optional, Tuple, Union
from uuid import UUID

from typing_extensions import Self

from camel.agents import ChatAgent


class CollectorData:
    def __init__(
        self,
        id: UUID,
        name: str,
        role: Literal["user", "assistant", "system", "tool"],
        message: Optional[str] = None,
        function_call: Optional[Dict[str, Any]] = None,
    ) -> None:
        r"""Create a data item store information about a message.
        Used by the data collector.

        Args:

            id (UUID): The id of the message.
            name (str): The name of the agent.
            role (Literal["user", "assistant", "system", "function"]):
                The role of the message.
            message (Optional[str], optional): The message.
                (default: :obj:`None`)
            function_call (Optional[Dict[str, Any]], optional):
                The function call. (default: :obj:`None`)

        Raises:

            ValueError: If the role is not supported.
            ValueError: If the role is system and function call is provided.
            ValueError: If neither message nor function call is provided.

        """
        if role not in ["user", "assistant", "system", "tool"]:
            raise ValueError(f"Role {role} not supported")
        if role == "system" and function_call:
            raise ValueError("System role cannot have function call")
        if not message and not function_call:
            raise ValueError(
                "Either message or function call must be provided"
            )
        self.id = id
        self.name = name
        self.role = role
        self.message = message
        self.function_call = function_call

    @staticmethod
    def from_context(name, context: Dict[str, Any]) -> "CollectorData":
        r"""Create a data collector from a context.

        Args:
            name (str): The name of the agent.
            context (Dict[str, Any]): The context.

        Returns:
            CollectorData: The data collector.
        """
        return CollectorData(
            id=uuid.uuid4(),
            name=name,
            role=context["role"],
            message=context["content"],
            function_call=context.get("tool_calls", None),
        )


class BaseDataCollector(ABC):
    r"""Base class for data collectors."""

    def __init__(self) -> None:
        r"""Create a data collector."""
        self.history: List[CollectorData] = []
        self._recording = False
        self.agents: List[Tuple[str, ChatAgent]] = []
        self.data: List[Dict[str, Any]] = []

    def step(
        self,
        role: Literal["user", "assistant", "system", "tool"],
        name: Optional[str] = None,
        message: Optional[str] = None,
        function_call: Optional[Dict[str, Any]] = None,
    ) -> Self:
        r"""Record a message.

        Args:
            role (Literal["user", "assistant", "system", "tool"]):
                The role of the message.
            name (Optional[str], optional): The name of the agent.
                (default: :obj:`None`)
            message (Optional[str], optional): The message to record.
                (default: :obj:`None`)
            function_call (Optional[Dict[str, Any]], optional):
                The function call to record. (default: :obj:`None`)

        Returns:
            Self: The data collector.

        """

        name = name or role

        self.history.append(
            CollectorData(
                id=uuid.uuid4(),
                name=name,
                role=role,
                message=message,
                function_call=function_call,
            )
        )
        return self

    def record(
        self,
        agent: Union[List[ChatAgent], ChatAgent],
    ) -> Self:
        r"""Record agents.

        Args:
            agent (Union[List[ChatAgent], ChatAgent]):
                The agent(s) to inject.
        """
        if not isinstance(agent, list):
            agent = [agent]
        for a in agent:
            name = a.role_name
            if not name:
                name = f"{a.__class__.__name__}_{len(self.agents)}"
            if name in [n for n, _ in self.agents]:
                raise ValueError(f"Name {name} already exists")

            self.agents.append((name, a))
        return self

    def start(self) -> Self:
        r"""Start recording."""
        self._recording = True
        return self

    def stop(self) -> Self:
        r"""Stop recording."""
        self._recording = False
        return self

    @property
    def recording(self) -> bool:
        r"""Whether the collector is recording."""
        return self._recording

    def reset(self, reset_agents: bool = True):
        r"""Reset the collector.

        Args:
            reset_agents (bool, optional):
                Whether to reset the agents. Defaults to True.
        """
        self.history = []
        if reset_agents:
            for _, agent in self.agents:
                agent.reset()

    @abstractmethod
    def convert(self) -> Any:
        r"""Convert the collected data."""
        pass

    @abstractmethod
    def llm_convert(self, converter: Any, prompt: Optional[str] = None) -> Any:
        r"""Convert the collected data."""
        pass

    def get_agent_history(self, name: str) -> List[CollectorData]:
        r"""Get the message history of an agent.

        Args:
            name (str): The name of the agent.

        Returns:
            List[CollectorData]: The message history of the agent
        """
        if not self.history:
            for _name, agent in self.agents:
                if _name == name:
                    return [
                        CollectorData.from_context(name, dict(i))
                        for i in agent.memory.get_context()[0]
                    ]
        return [msg for msg in self.history if msg.name == name]
